__author__ = "sibirrer"

import numpy.testing as npt
import pytest
import numpy as np

from lenstronomy.Analysis.td_cosmography import TDCosmography
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.Util import constants as const


class TestTDCosmography(object):
    def setup_method(self):
        kwargs_model = {
            "lens_light_model_list": ["HERNQUIST"],
            "lens_model_list": ["SIE"],
            "point_source_model_list": ["LENSED_POSITION"],
        }
        z_lens = 0.5
        z_source = 2.5

        R_slit = 3.8
        dR_slit = 1.0
        aperture_type = "slit"
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "center_ra": 0,
            "width": dR_slit,
            "length": R_slit,
            "angle": 0,
            "center_dec": 0,
        }
        psf_fwhm = 0.7
        kwargs_seeing = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}

        TDCosmography(z_lens, z_source, kwargs_model)
        from astropy.cosmology import FlatLambdaCDM

        cosmo = FlatLambdaCDM(H0=70, Om0=0.3, Ob0=0.05)
        self.td_cosmo = TDCosmography(
            z_lens,
            z_source,
            kwargs_model,
            cosmo_fiducial=cosmo,
            lens_model_kinematics_bool=None,
            kwargs_aperture=kwargs_aperture,
            kwargs_seeing=kwargs_seeing,
            light_model_kinematics_bool=None,
        )
        self.lens = LensModel(
            lens_model_list=["SIE"], cosmo=cosmo, z_lens=z_lens, z_source=z_source
        )
        self.solver = LensEquationSolver(lensModel=self.lens)

        self.kwargs_lens = [
            {"theta_E": 1, "e1": 0.1, "e2": -0.2, "center_x": 0, "center_y": 0}
        ]
        source_x, source_y = 0, 0.05
        image_x, image_y = self.solver.image_position_from_source(
            source_x, source_y, self.kwargs_lens, min_distance=0.1, search_window=10
        )
        self.kwargs_ps = [{"ra_image": image_x, "dec_image": image_y}]
        self.image_x, self.image_y = image_x, image_y

    def test_time_delays(self):
        dt = self.td_cosmo.time_delays(self.kwargs_lens, self.kwargs_ps, kappa_ext=0)
        dt_true = self.lens.arrival_time(self.image_x, self.image_y, self.kwargs_lens)
        npt.assert_almost_equal(dt, dt_true, decimal=6)

    def test_fermat_potential(self):
        fermat_pot = self.td_cosmo.fermat_potential(self.kwargs_lens, self.kwargs_ps)
        fermat_pot_true = self.lens.fermat_potential(
            self.image_x, self.image_y, self.kwargs_lens
        )
        npt.assert_almost_equal(fermat_pot, fermat_pot_true, decimal=6)

        diff = 0.1
        kwargs_ps = [{"ra_image": self.image_x + diff, "dec_image": self.image_y}]
        fermat_pot = self.td_cosmo.fermat_potential(self.kwargs_lens, kwargs_ps)
        fermat_pot_true = self.lens.fermat_potential(
            self.image_x + diff, self.image_y, self.kwargs_lens
        )
        ratio = fermat_pot / fermat_pot_true
        assert np.max(np.abs(ratio)) > 1.05

    def test_cosmo_inference(self):
        # set up a cosmology
        # compute image postions
        # compute J and velocity dispersion
        D_dt = self.td_cosmo._lens_cosmo.ddt
        D_d = self.td_cosmo._lens_cosmo.dd
        D_s = self.td_cosmo._lens_cosmo.ds
        D_ds = self.td_cosmo._lens_cosmo.dds
        fermat_potential_list = self.td_cosmo.fermat_potential(
            self.kwargs_lens, self.kwargs_ps
        )
        dt_list = self.td_cosmo.time_delays(
            self.kwargs_lens, self.kwargs_ps, kappa_ext=0
        )
        dt = dt_list[0] - dt_list[1]
        d_fermat = fermat_potential_list[0] - fermat_potential_list[1]

        D_dt_infered = self.td_cosmo.ddt_from_time_delay(
            d_fermat_model=d_fermat, dt_measured=dt
        )
        npt.assert_almost_equal(D_dt_infered, D_dt, decimal=5)
        r_eff = 0.5
        kwargs_lens_light = [{"Rs": r_eff * 0.551, "center_x": 0, "center_y": 0}]
        kwargs_anisotropy = {"r_ani": 1}

        anisotropy_model = "OM"
        kwargs_numerics_galkin = {
            "interpol_grid_num": 500,
            "log_integration": True,
            "max_integrate": 10,
            "min_integrate": 0.001,
        }
        self.td_cosmo.kinematics_modeling_settings(
            anisotropy_model,
            kwargs_numerics_galkin,
            analytic_kinematics=True,
            Hernquist_approx=False,
            MGE_light=False,
            MGE_mass=False,
        )

        J = self.td_cosmo.velocity_dispersion_dimension_less(
            self.kwargs_lens,
            kwargs_lens_light,
            kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=self.kwargs_lens[0]["theta_E"],
            gamma=2,
        )

        J_map = self.td_cosmo.velocity_dispersion_map_dimension_less(
            self.kwargs_lens,
            kwargs_lens_light,
            kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=self.kwargs_lens[0]["theta_E"],
            gamma=2,
        )
        assert len(J_map) == 1
        npt.assert_almost_equal(J_map[0] / J, 1, decimal=1)
        sigma_v2 = J * D_s / D_ds * const.c**2
        sigma_v = np.sqrt(sigma_v2) / 1000.0  # convert to [km/s]
        print(sigma_v, "test sigma_v")
        Ds_Dds = self.td_cosmo.ds_dds_from_kinematics(sigma_v, J, kappa_s=0, kappa_ds=0)
        npt.assert_almost_equal(Ds_Dds, D_s / D_ds)

        # now we perform a mass-sheet transform in the observables but leave the models identical with a convergence correction
        kappa_s = 0.5
        dt_list = self.td_cosmo.time_delays(
            self.kwargs_lens, self.kwargs_ps, kappa_ext=kappa_s
        )
        sigma_v_kappa = sigma_v * np.sqrt(1 - kappa_s)
        dt = dt_list[0] - dt_list[1]
        D_dt_infered, D_d_infered = self.td_cosmo.ddt_dd_from_time_delay_and_kinematics(
            d_fermat_model=d_fermat,
            dt_measured=dt,
            sigma_v_measured=sigma_v_kappa,
            J=J,
            kappa_s=kappa_s,
            kappa_ds=0,
            kappa_d=0,
        )
        npt.assert_almost_equal(D_dt_infered, D_dt, decimal=6)
        npt.assert_almost_equal(D_d_infered, D_d, decimal=6)


if __name__ == "__main__":
    pytest.main()



__author__ = "jhodonnell"


import numpy as np
import numpy.testing as npt
import unittest
import pytest

from lenstronomy.Sampling.param_group import ModelParamGroup, SingleParam, ArrayParam


class ExampleSingleParam(SingleParam):
    param_names = ["sp1", "sp2"]
    _kwargs_lower = {"sp1": 0, "sp2": 0}
    _kwargs_upper = {"sp1": 10, "sp2": 10}


class ExampleArrayParam(ArrayParam):
    param_names = {"ap1": 1, "ap2": 3}
    _kwargs_lower = {"ap1": [0], "ap2": [0] * 3}
    _kwargs_upper = {"ap1": [10], "ap2": [10] * 3}


class TestParamGroup(object):
    def setup_method(self):
        pass

    def test_single_param(self):
        sp = ExampleSingleParam(on=True)

        num, names = sp.num_params({})
        assert num == 2
        assert names == ["sp1", "sp2"]

        result = sp.set_params({"sp1": 2}, {"sp2": 3})
        assert result == [2]

        result = sp.set_params({"sp1": 2, "sp2": 3}, {})
        assert result == [2, 3]

        kwargs, i = sp.get_params(result, i=0, kwargs_fixed={})
        assert kwargs["sp1"] == 2
        assert kwargs["sp2"] == 3

    def test_array_param(self):
        ap = ExampleArrayParam(on=True)

        num, names = ap.num_params({})
        assert num == 4
        assert names == ["ap1"] + ["ap2"] * 3

        result = ap.set_params({"ap1": [2]}, {"ap2": [1, 1, 1]})
        assert result == [2]

        result = ap.set_params({"ap1": [2], "ap2": [1, 2, 3]}, {})
        assert result == [2, 1, 2, 3]

        kwargs, i = ap.get_params(result, i=0, kwargs_fixed={})
        assert kwargs["ap1"] == [2]
        assert kwargs["ap2"] == [1, 2, 3]

    def test_compose(self):
        sp = ExampleSingleParam(on=True)
        ap = ExampleArrayParam(on=True)

        num, names = ModelParamGroup.compose_num_params([sp, ap], kwargs_fixed={})
        assert num == 6
        assert names == sp.num_params({})[1] + ap.num_params({})[1]

        result = ModelParamGroup.compose_set_params(
            [sp, ap],
            {"sp1": 1, "sp2": 2, "ap1": [3], "ap2": [4, 5, 6]},
            kwargs_fixed={},
        )
        assert result == [1, 2, 3, 4, 5, 6]

        kwargs, i = ModelParamGroup.compose_get_params(
            [sp, ap], result, i=0, kwargs_fixed={}
        )
        assert kwargs["sp1"] == 1
        assert kwargs["ap2"] == [4, 5, 6]


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"
# this file contains a class to make a gaussian

import numpy as np
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["Gaussian"]


class Gaussian(LensProfileBase):
    """This class contains functions to evaluate a Gaussian function and calculates its
    derivative and hessian matrix."""

    param_names = ["amp", "sigma_x", "sigma_y", "center_x", "center_y"]
    lower_limit_default = {"amp": 0, "sigma": 0, "center_x": -100, "center_y": -100}
    upper_limit_default = {"amp": 100, "sigma": 100, "center_x": 100, "center_y": 100}

    def function(self, x, y, amp, sigma_x, sigma_y, center_x=0, center_y=0):
        """Returns Gaussian."""
        c = amp / (2 * np.pi * sigma_x * sigma_y)
        delta_x = x - center_x
        delta_y = y - center_y
        exponent = -((delta_x / sigma_x) ** 2 + (delta_y / sigma_y) ** 2) / 2.0
        return c * np.exp(exponent)

    def derivatives(self, x, y, amp, sigma_x, sigma_y, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function."""
        f_ = self.function(x, y, amp, sigma_x, sigma_y, center_x, center_y)
        return f_ * (center_x - x) / sigma_x**2, f_ * (center_y - y) / sigma_y**2

    def hessian(self, x, y, amp, sigma_x, sigma_y, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx,
        d^f/dy^2."""
        f_ = self.function(x, y, amp, sigma_x, sigma_y, center_x, center_y)
        f_xx = f_ * ((-1.0 / sigma_x**2) + (center_x - x) ** 2 / sigma_x**4)
        f_yy = f_ * ((-1.0 / sigma_y**2) + (center_y - y) ** 2 / sigma_y**4)
        f_xy = f_ * (center_x - x) / sigma_x**2 * (center_y - y) / sigma_y**2
        return f_xx, f_xy, f_xy, f_yy



import numpy as np
from lenstronomy.Util import class_creator

__all__ = ["ImageLikelihood"]


class ImageLikelihood(object):
    """Manages imaging data likelihoods."""

    def __init__(
        self,
        multi_band_list,
        multi_band_type,
        kwargs_model,
        bands_compute=None,
        image_likelihood_mask_list=None,
        source_marg=False,
        linear_prior=None,
        check_positive_flux=False,
        kwargs_pixelbased=None,
        linear_solver=True,
    ):
        """

        :param bands_compute: list of bools with same length as data objects, indicates which "band" to include in the
         fitting
        :param image_likelihood_mask_list: list of boolean 2d arrays of size of images marking the pixels to be
         evaluated in the likelihood
        :param source_marg: marginalization addition on the imaging likelihood based on the covariance of the inferred
         linear coefficients
        :param linear_prior: float or list of floats (when multi-linear setting is chosen) indicating the range of
         linear amplitude priors when computing the marginalization term.
        :param check_positive_flux: bool, option to punish models that do not have all positive linear amplitude
         parameters
        :param kwargs_pixelbased: keyword arguments with various settings related to the pixel-based solver
         (see SLITronomy documentation)
        :param linear_solver: bool, if True (default) fixes the linear amplitude parameters 'amp' (avoid sampling) such
         that they get overwritten by the linear solver solution.
        """
        self.imSim = class_creator.create_im_sim(
            multi_band_list,
            multi_band_type,
            kwargs_model,
            bands_compute=bands_compute,
            image_likelihood_mask_list=image_likelihood_mask_list,
            kwargs_pixelbased=kwargs_pixelbased,
            linear_solver=linear_solver,
        )
        self._model_type = self.imSim.type
        self._source_marg = source_marg
        self._linear_prior = linear_prior
        self._check_positive_flux = check_positive_flux

    def logL(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_special=None,
        kwargs_extinction=None,
    ):
        """

        :param kwargs_lens: lens model keyword argument list according to LensModel module
        :param kwargs_source: source light keyword argument list according to LightModel module
        :param kwargs_lens_light: deflector light (not lensed) keyword argument list according to LightModel module
        :param kwargs_ps: point source keyword argument list according to PointSource module
        :param kwargs_special: special keyword argument list as part of the Param module
        :param kwargs_extinction: extinction parameter keyword argument list according to LightModel module
        :return: log likelihood of the data given the model
        """
        logL = self.imSim.likelihood_data_given_model(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light,
            kwargs_ps,
            kwargs_extinction=kwargs_extinction,
            kwargs_special=kwargs_special,
            source_marg=self._source_marg,
            linear_prior=self._linear_prior,
            check_positive_flux=self._check_positive_flux,
        )
        if np.isnan(logL) is True:
            return -(10**15)
        return logL

    @property
    def num_data(self):
        """

        :return: number of image data points
        """
        return self.imSim.num_data_evaluate

    def num_param_linear(
        self,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_special=None,
        kwargs_extinction=None,
    ):
        """

        :return:  number of linear parameters solved for during the image reconstruction process
        """
        return self.imSim.num_param_linear(
            kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps
        )

    def reset_point_source_cache(self, cache=True):
        """

        :param cache: boolean
        :return: None
        """
        self.imSim.reset_point_source_cache(cache=cache)



__author__ = "dgilman"

from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["TabulatedDeflections"]


class TabulatedDeflections(LensProfileBase):
    """A user-defined class that returns deflection angles given a set of observed
    coordinates on the sky (x, y).

    This class has similar functionality as INTERPOL, with the difference being that the
    interpolation for this class is done prior to class creation. When used with
    routines in the lenstronomy.Sampling, this class effectively acts as a fixed lens
    model with no keyword arguments.
    """

    profile_name = "TABULATED_DEFLECTIONS"
    param_names = []
    lower_limit_default = {}
    upper_limit_default = {}

    def __init__(self, custom_class):
        """
        :param custom_class: a user-defined class that has a __call___ method that returns deflection angles

        Code example:

        >>> custom_class = CustomLensingClass()
        >>> alpha_x, alpha_y = custom_class(x, y, **kwargs)

        or equivalently:

        >>> from lenstronomy.LensModel.lens_model import LensModel
        >>> lens_model_list = ['NumericalAlpha']
        >>> lens_model = LensModel(lens_model_list, numerical_alpha_class=custom_class)
        >>>> alpha_x, alpha_y = lens_model.alpha(x, y, **kwargs)
        """

        self._custom_lens_class = custom_class
        super(TabulatedDeflections, self).__init__()

    def function(self, x, y, center_x=0, center_y=0, **kwargs):
        raise Exception("no potential for this class.")

    def derivatives(self, x, y, center_x=0, center_y=0, **kwargs):
        """

        :param x: x coordinate [arcsec]
        :param y: x coordinate [arcsec]
        :param center_x: deflector x center [arcsec]
        :param center_y: deflector y center [arcsec]
        :param kwargs: keyword arguments for the custom profile
        :return:
        """

        x_ = x - center_x
        y_ = y - center_y
        f_x, f_y = self._custom_lens_class(x_, y_, **kwargs)

        return f_x, f_y

    def hessian(self, x, y, center_x=0, center_y=0, **kwargs):
        """Returns the components of the hessian matrix :param x: x coordinate [arcsec]
        :param y: y coordinate [arcsec] :param center_x: the deflector x coordinate
        :param center_y: the deflector y coordinate :param kwargs: keyword arguments for
        the profile :return: the derivatives of the deflection angles that make up the
        hessian matrix."""

        diff = 1e-6
        alpha_ra, alpha_dec = self.derivatives(
            x, y, center_x=center_x, center_y=center_y, **kwargs
        )

        alpha_ra_dx, alpha_dec_dx = self.derivatives(
            x + diff, y, center_x=center_x, center_y=center_y, **kwargs
        )
        alpha_ra_dy, alpha_dec_dy = self.derivatives(
            x, y + diff, center_x=center_x, center_y=center_y, **kwargs
        )

        dalpha_rara = (alpha_ra_dx - alpha_ra) / diff
        dalpha_radec = (alpha_ra_dy - alpha_ra) / diff
        dalpha_decra = (alpha_dec_dx - alpha_dec) / diff
        dalpha_decdec = (alpha_dec_dy - alpha_dec) / diff

        f_xx = dalpha_rara
        f_yy = dalpha_decdec
        f_xy = dalpha_radec
        f_yx = dalpha_decra

        return f_xx, f_xy, f_yx, f_yy



__author__ = "sibirrer"

import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
from lenstronomy.LensModel.Profiles.convergence import Convergence
import numpy as np

__all__ = ["Shear", "ShearGammaPsi", "ShearReduced"]


class Shear(LensProfileBase):
    """Class for external shear gamma1, gamma2 expression."""

    param_names = ["gamma1", "gamma2", "ra_0", "dec_0"]
    lower_limit_default = {"gamma1": -0.5, "gamma2": -0.5, "ra_0": -100, "dec_0": -100}
    upper_limit_default = {"gamma1": 0.5, "gamma2": 0.5, "ra_0": 100, "dec_0": 100}

    def function(self, x, y, gamma1, gamma2, ra_0=0, dec_0=0):
        """

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param gamma1: shear component
        :param gamma2: shear component
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return: lensing potential
        """
        x_ = x - ra_0
        y_ = y - dec_0
        f_ = 1 / 2.0 * (gamma1 * x_ * x_ + 2 * gamma2 * x_ * y_ - gamma1 * y_ * y_)
        return f_

    def derivatives(self, x, y, gamma1, gamma2, ra_0=0, dec_0=0):
        """

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param gamma1: shear component
        :param gamma2: shear component
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return: deflection angles
        """
        x_ = x - ra_0
        y_ = y - dec_0
        f_x = gamma1 * x_ + gamma2 * y_
        f_y = +gamma2 * x_ - gamma1 * y_
        return f_x, f_y

    def hessian(self, x, y, gamma1, gamma2, ra_0=0, dec_0=0):
        """

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param gamma1: shear component
        :param gamma2: shear component
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return: f_xx, f_xy, f_yx, f_yy
        """
        gamma1 = gamma1
        gamma2 = gamma2
        kappa = 0
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy


class ShearGammaPsi(LensProfileBase):
    """
    class to model a shear field with shear strength and direction. The translation ot the cartesian shear distortions
    is as follow:

    .. math::
        \\gamma_1 = \\gamma_{ext} \\cos(2 \\phi_{ext})
        \\gamma_2 = \\gamma_{ext} \\sin(2 \\phi_{ext})

    """

    param_names = ["gamma_ext", "psi_ext", "ra_0", "dec_0"]
    lower_limit_default = {
        "gamma_ext": 0,
        "psi_ext": -np.pi,
        "ra_0": -100,
        "dec_0": -100,
    }
    upper_limit_default = {"gamma_ext": 1, "psi_ext": np.pi, "ra_0": 100, "dec_0": 100}

    def __init__(self):
        self._shear_e1e2 = Shear()
        super(ShearGammaPsi, self).__init__()

    @staticmethod
    def function(x, y, gamma_ext, psi_ext, ra_0=0, dec_0=0):
        """

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param gamma_ext: shear strength
        :param psi_ext: shear angle (radian)
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return:
        """
        # change to polar coordinate
        r, phi = param_util.cart2polar(x - ra_0, y - dec_0)
        f_ = 1.0 / 2 * gamma_ext * r**2 * np.cos(2 * (phi - psi_ext))
        return f_

    def derivatives(self, x, y, gamma_ext, psi_ext, ra_0=0, dec_0=0):
        # rotation angle
        gamma1, gamma2 = param_util.shear_polar2cartesian(psi_ext, gamma_ext)
        return self._shear_e1e2.derivatives(x, y, gamma1, gamma2, ra_0, dec_0)

    def hessian(self, x, y, gamma_ext, psi_ext, ra_0=0, dec_0=0):
        gamma1, gamma2 = param_util.shear_polar2cartesian(psi_ext, gamma_ext)
        return self._shear_e1e2.hessian(x, y, gamma1, gamma2, ra_0, dec_0)


class ShearReduced(LensProfileBase):
    """Reduced shear distortions :math:`\\gamma' = \\gamma / (1-\\kappa)`. This
    distortion keeps the magnification as unity and, thus, does not change the size of
    apparent objects. To keep the magnification at unity, it requires.

    .. math::
        (1-\\kappa)^2) - \\gamma_1^2 - \\gamma_2^ = 1

    Thus, for given pair of reduced shear :math:`(\\gamma'_1, \\gamma'_2)`, an additional convergence term is calculated
    and added to the lensing distortions.
    """

    param_names = ["gamma1", "gamma2", "ra_0", "dec_0"]
    lower_limit_default = {"gamma1": -0.5, "gamma2": -0.5, "ra_0": -100, "dec_0": -100}
    upper_limit_default = {"gamma1": 0.5, "gamma2": 0.5, "ra_0": 100, "dec_0": 100}

    def __init__(self):
        self._shear = Shear()
        self._convergence = Convergence()
        super(ShearReduced, self).__init__()

    @staticmethod
    def _kappa_reduced(gamma1, gamma2):
        """Compute convergence such that magnification is unity.

        :param gamma1: reduced shear
        :param gamma2: reduced shear
        :return: kappa
        """
        kappa = 1 - 1.0 / np.sqrt(1 - gamma1**2 - gamma2**2)
        gamma1_ = (1 - kappa) * gamma1
        gamma2_ = (1 - kappa) * gamma2
        return kappa, gamma1_, gamma2_

    def function(self, x, y, gamma1, gamma2, ra_0=0, dec_0=0):
        """

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param gamma1: shear component
        :param gamma2: shear component
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return: lensing potential
        """
        kappa, gamma1_, gamma2_ = self._kappa_reduced(gamma1, gamma2)
        f_shear = self._shear.function(x, y, gamma1_, gamma2_, ra_0, dec_0)
        f_kappa = self._convergence.function(x, y, kappa, ra_0, dec_0)
        return f_shear + f_kappa

    def derivatives(self, x, y, gamma1, gamma2, ra_0=0, dec_0=0):
        """

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param gamma1: shear component
        :param gamma2: shear component
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return: deflection angles
        """
        kappa, gamma1_, gamma2_ = self._kappa_reduced(gamma1, gamma2)
        f_x_shear, f_y_shear = self._shear.derivatives(
            x, y, gamma1_, gamma2_, ra_0, dec_0
        )
        f_x_kappa, f_y_kappa = self._convergence.derivatives(x, y, kappa, ra_0, dec_0)
        return f_x_shear + f_x_kappa, f_y_shear + f_y_kappa

    def hessian(self, x, y, gamma1, gamma2, ra_0=0, dec_0=0):
        """

        :param x: x-coordinate (angle)
        :param y: y0-coordinate (angle)
        :param gamma1: shear component
        :param gamma2: shear component
        :param ra_0: x/ra position where shear deflection is 0
        :param dec_0: y/dec position where shear deflection is 0
        :return: f_xx, f_xy, f_yx, f_yy
        """
        kappa, gamma1_, gamma2_ = self._kappa_reduced(gamma1, gamma2)
        f_xx_g, f_xy_g, f_yx_g, f_yy_g = self._shear.hessian(
            x, y, gamma1_, gamma2_, ra_0, dec_0
        )
        f_xx_k, f_xy_k, f_yx_k, f_yy_k = self._convergence.hessian(
            x, y, kappa, ra_0, dec_0
        )
        f_xx = f_xx_g + f_xx_k
        f_yy = f_yy_g + f_yy_k
        f_xy = f_xy_g + f_xy_k
        return f_xx, f_xy, f_xy, f_yy



__author__ = "lucateo"

import numpy as np
from scipy.special import exp1, erf
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["CoredDensityExp"]


class CoredDensityExp(LensProfileBase):
    """This class contains functions concerning an exponential cored density profile,
    namely.

    ..math::
        \\rho(r) = \\rho_0 \\exp(- (\\theta / \\theta_c)^2)
    """

    _s = 0.000001  # numerical limit for minimal radius
    param_names = ["kappa_0", "theta_c", "center_x", "center_y"]
    lower_limit_default = {
        "kappa_0": 0,
        "theta_c": 0,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "kappa_0": 10,
        "theta_c": 100,
        "center_x": 100,
        "center_y": 100,
    }

    @staticmethod
    def rhotilde(kappa_0, theta_c):
        """Computes the central density in angular units :param kappa_0: central
        convergence of profile :param theta_c: core radius (in arcsec) :return: central
        density in 1/arcsec."""
        return kappa_0 / (np.sqrt(np.pi) * theta_c)

    def function(self, x, y, kappa_0, theta_c, center_x=0, center_y=0):
        """
        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: lensing potential (in arcsec^2)
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        r = np.maximum(r, self._s)
        Integral_factor = 0.5 * exp1((r / theta_c) ** 2) + np.log((r / theta_c))
        function = kappa_0 * theta_c**2 * Integral_factor
        return function

    @staticmethod
    def alpha_radial(r, kappa_0, theta_c):
        """Returns the radial part of the deflection angle :param r: angular position
        (normally in units of arc seconds) :param kappa_0: central convergence of
        profile :param theta_c: core radius (in arcsec) :return: radial deflection
        angle."""
        prefactor = kappa_0 * theta_c**2 / r
        return prefactor * (1 - np.exp(-((r / theta_c) ** 2)))

    def derivatives(self, x, y, kappa_0, theta_c, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function (lensing potential), which are the
        deflection angles.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: deflection angle in x, deflection angle in y
        """
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        R = np.maximum(R, 0.00000001)
        f_x = self.alpha_radial(R, kappa_0, theta_c) * x_ / R
        f_y = self.alpha_radial(R, kappa_0, theta_c) * y_ / R
        return f_x, f_y

    def hessian(self, x, y, kappa_0, theta_c, center_x=0, center_y=0):
        """
        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx, d^f/dy^2
        """
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        R = np.maximum(R, 0.00000001)
        prefactor = kappa_0 * theta_c**2
        expFactor = np.exp(-((R / theta_c) ** 2))
        factor1 = (1 - expFactor) / R**4
        factor2 = 2 / (R**2 * theta_c**2) * expFactor
        f_xx = prefactor * (factor1 * (y_**2 - x_**2) + factor2 * x_**2)
        f_yy = prefactor * (factor1 * (x_**2 - y_**2) + factor2 * y_**2)
        f_xy = prefactor * (-factor1 * 2 * x_ * y_ + factor2 * x_ * y_)
        return f_xx, f_xy, f_xy, f_yy

    def density(self, R, kappa_0, theta_c):
        """Three dimensional density profile in angular units (rho0_physical =
        rho0_angular Sigma_crit / D_lens)

        :param R: projected angular position (normally in units of arc seconds)
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :return: rho(R) density
        """
        rhotilde = self.rhotilde(kappa_0, theta_c)
        return rhotilde * np.exp(-((R / theta_c) ** 2))

    def density_lens(self, r, kappa_0, theta_c):
        """Computes the density at 3d radius r given lens model parameterization. The
        integral in the LOS projection of this quantity results in the convergence
        quantity.

        :param r: angular position (normally in units of arc seconds)
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :return: density rho(r)
        """
        return self.density(r, kappa_0, theta_c)

    @staticmethod
    def kappa_r(R, kappa_0, theta_c):
        """Convergence of the cored density profile. This routine is also for testing.

        :param R: radius (angular scale)
        :param kappa_0: convergence in the core
        :param theta_c: core radius
        :return: convergence at r
        """
        expFactor = np.exp(-((R / theta_c) ** 2))
        return kappa_0 * expFactor

    def density_2d(self, x, y, kappa_0, theta_c, center_x=0, center_y=0):
        """Projected two dimensional ULDM profile (convergence * Sigma_crit), but given
        our units convention for rho0, it is basically the convergence.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcsec)
        :return: Epsilon(R) projected density at radius R
        """
        x_ = x - center_x
        y_ = y - center_y
        R = np.sqrt(x_**2 + y_**2)
        return self.kappa_r(R, kappa_0, theta_c)

    @staticmethod
    def mass_3d(R, kappa_0, theta_c):
        """Mass enclosed a 3d sphere or radius r :param kappa_0: central convergence of
        profile :param theta_c: core radius (in arcsec) :param R: radius in arcseconds
        :return: mass of soliton in angular units."""
        integral_factor = np.sqrt(np.pi) * erf(R / theta_c) / 2 - R / theta_c * np.exp(
            -((R / theta_c) ** 2)
        )
        m_3d = 2 * np.sqrt(np.pi) * kappa_0 * theta_c**2 * integral_factor
        return m_3d

    def mass_3d_lens(self, r, kappa_0, theta_c):
        """Mass enclosed a 3d sphere or radius r :param kappa_0: central convergence of
        profile :param theta_c: core radius (in arcsec) :return: mass."""
        m_3d = self.mass_3d(r, kappa_0, theta_c)
        return m_3d

    def mass_2d(self, R, kappa_0, theta_c):
        """Mass enclosed a 2d sphere of radius r returns.

        .. math::
            M_{2D} = 2 \\pi \\int_0^r dr' r' \\int dz \\rho(\\sqrt(r'^2 + z^2))

        :param kappa_0: central convergence of soliton
        :param theta_c: core radius (in arcsec)
        :return: M_2D (ULDM only)
        """
        return self.alpha_radial(R, kappa_0, theta_c) * np.pi * R



import numpy as np
from lenstronomy.GalKin.galkin import Galkin
from lenstronomy.Util import util
from lenstronomy.Util import mask_util
from scipy import signal


class GalkinShells(Galkin):
    """Class to calculate velocity dispersion for radial shells in a fast way."""

    def __init__(
        self,
        kwargs_model,
        kwargs_aperture,
        kwargs_psf,
        kwargs_cosmo,
        kwargs_numerics=None,
        analytic_kinematics=False,
    ):
        """

        :param kwargs_model: keyword arguments describing the model components
        :param kwargs_aperture: keyword arguments describing the spectroscopic aperture, see Aperture() class
        :param kwargs_psf: keyword argument specifying the PSF of the observation
        :param kwargs_cosmo: keyword arguments that define the cosmology in terms of the angular diameter distances
         involved
        :param kwargs_numerics: numerics keyword arguments
        :param analytic_kinematics: bool, if True uses the analytic kinematic model
        """
        Galkin.__init__(
            self,
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics,
            analytic_kinematics=analytic_kinematics,
        )
        if not self.aperture_type == "IFU_shells":
            raise ValueError(
                'GalkinShells is not supported with aperture_type %s. Only support with "IFU_shells"'
                % self.aperture_type
            )
        self._r_bins = kwargs_aperture["r_bins"]
        r_max = np.max(self._r_bins)
        self._num_pix = 100
        # factor of 1.5 to allow outside flux to be convolved into the largest bin
        self._delta_pix = 1.5 * r_max * 2 / self._num_pix
        x_grid, y_grid = util.make_grid(numPix=self._num_pix, deltapix=self._delta_pix)
        self._r_grid = np.sqrt(x_grid**2 + y_grid**2)

    def dispersion_map(self, kwargs_mass, kwargs_light, kwargs_anisotropy, **kwargs):
        """

        :param kwargs_mass: mass model parameters (following lenstronomy lens model conventions)
        :param kwargs_light: deflector light parameters (following lenstronomy light model conventions)
        :param kwargs_anisotropy: anisotropy parameters, may vary according to anisotropy type chosen.
         We refer to the Anisotropy() class for details on the parameters.
        :return: array of velocity dispersion for each IFU shell [km/s]
        """
        I_R_sigma2, IR = self.numerics._I_R_sigma2_interp(
            self._r_grid, kwargs_mass, kwargs_light, kwargs_anisotropy
        )
        ir_map = util.array2image(IR)
        ir_sigma2_map = util.array2image(I_R_sigma2)
        kernel = self.convolution_kernel(
            delta_pix=self._delta_pix, num_pix=self._num_pix
        )
        I_R_sigma2_conv = signal.fftconvolve(ir_sigma2_map, kernel, mode="same")
        I_R_sigma2_conv = util.image2array(I_R_sigma2_conv)
        I_R_conv = signal.fftconvolve(ir_map, kernel, mode="same")
        I_R_conv = util.image2array(I_R_conv)

        # average over radial bins
        vel_disp_array = []
        r_min = self._r_bins[0]
        for r_max in self._r_bins[1:]:
            mask = mask_util.mask_shell(
                self._r_grid, 0, center_x=0, center_y=0, r_in=r_min, r_out=r_max
            )
            vel_disp = np.sum(I_R_sigma2_conv * mask) / np.sum(
                I_R_conv * mask
            )  # luminosity weighted average
            vel_disp_array.append(vel_disp)
            r_min = r_max
        self.numerics.delete_cache()
        return np.sqrt(np.array(vel_disp_array)) / 1000



__author__ = "aymgal, johannesulf"

import lenstronomy.Util.sampling_util as utils
import numpy as np

__all__ = ["NestedSampler"]


class NestedSampler(object):
    """Base class for nested samplers."""

    def __init__(
        self,
        likelihood_module,
        prior_type,
        prior_means,
        prior_sigmas,
        width_scale,
        sigma_scale,
    ):
        """
        :param likelihood_module: likelihood_module like in likelihood.py (should be callable)
        :param prior_type: 'uniform' of 'gaussian', for converting the unit hypercube to param cube
        :param prior_means: if prior_type is 'gaussian', mean for each param
        :param prior_sigmas: if prior_type is 'gaussian', std dev for each param
        :param width_scale: scale the widths of the parameters space by this factor
        :param sigma_scale: if prior_type is 'gaussian', scale the gaussian sigma by this factor
        """
        self._ll = likelihood_module
        self.n_dims, self.param_names = self._ll.param.num_param()

        lowers, uppers = self._ll.param_limits
        if width_scale < 1:
            self.lowers, self.uppers = utils.scale_limits(lowers, uppers, width_scale)
        else:
            self.lowers, self.uppers = lowers, uppers

        if prior_type == "gaussian":
            if prior_means is None or prior_sigmas is None:
                raise ValueError(
                    "For gaussian prior type, means and sigmas are required"
                )
            self.means, self.sigmas = prior_means, prior_sigmas * sigma_scale
            self.lowers, self.uppers = lowers, uppers
        elif prior_type != "uniform":
            raise ValueError("Sampling type {} not supported".format(prior_type))
        self.prior_type = prior_type
        self._has_warned = False

    def prior(self, u, *args):
        """Compute the mapping between the unit cube and parameter cube.

        :param u: unit hypercube, sampled by the algorithm
        :return: hypercube in parameter space
        """

        # MultiNest passes its own type. Make a copy of the original and create
        # a Python version.
        u_orig = u
        u = np.array([u[i] for i in range(self.n_dims)])

        if self.prior_type == "gaussian":
            p = utils.cube2args_gaussian(
                u,
                self.lowers,
                self.uppers,
                self.means,
                self.sigmas,
                self.n_dims,
                copy=True,
            )
        elif self.prior_type == "uniform":
            p = utils.cube2args_uniform(
                u, self.lowers, self.uppers, self.n_dims, copy=True
            )
        else:
            raise ValueError(
                'prior type %s not supported! Chose "gaussian" or "uniform".'
            )

        # MultiNest expects that we modify the origal array instead of
        # returning the transformed parameters.
        for i in range(self.n_dims):
            # But for PolyChord, this is read-only.
            try:
                u_orig[i] = p[i]
            except ValueError:
                pass

        return p

    def log_likelihood(self, p, *args):
        """Compute the log-likelihood given list of parameters.

        :param x: parameter values
        :return: log-likelihood (from the likelihood module)
        """

        # MultiNest passes its own type.
        p = np.array([p[i] for i in range(self.n_dims)])

        log_l = self._ll(p)
        if not np.isfinite(log_l):
            if not self._has_warned:
                print("WARNING : logL is not finite : return very low value instead")
            log_l = -1e15
            self._has_warned = True
        return float(log_l)

    def run(self, kwargs_run):
        """Run the nested sampling algorithm."""
        raise NotImplementedError("Method not be implemented in base class")



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
import pytest
from lenstronomy.LensModel.Profiles.curved_arc_sis_mst import CurvedArcSISMST
from lenstronomy.LensModel.Profiles.curved_arc_const import (
    CurvedArcConstMST,
    CurvedArcConst,
)
from lenstronomy.Util import util


class TestCurvedArcConst(object):
    """Tests the source model routines."""

    def setup_method(self):
        self.arc_sis = CurvedArcSISMST()
        self.arc_const = CurvedArcConst()

    def test_function(self):
        kwargs_arc = {
            "tangential_stretch": 5,
            #'radial_stretch': 1,
            "curvature": 1.0 / 10,
            "direction": 0,
            "center_x": 0,
            "center_y": 0,
        }
        npt.assert_raises(Exception, self.arc_const.function, 0.0, 0.0, **kwargs_arc)

    def test_derivatives(self):
        kwargs_arc = {
            "tangential_stretch": 3,
            #'radial_stretch': 1.,
            "curvature": 0.8,
            "direction": 0,
            "center_x": 0,
            "center_y": 0,
        }

        kwargs_arc_sis = {
            "tangential_stretch": 3,
            "radial_stretch": 1.0,
            "curvature": 0.8,
            "direction": 0,
            "center_x": 0,
            "center_y": 0,
        }
        x, y = util.make_grid(numPix=100, deltapix=0.01)
        f_x_sis, f_y_sis = self.arc_sis.derivatives(x, y, **kwargs_arc_sis)
        beta_x_sis = x - f_x_sis
        beta_y_sis = y - f_y_sis
        f_x_const, f_y_const = self.arc_const.derivatives(x, y, **kwargs_arc)
        beta_x_const = x - f_x_const
        beta_y_const = y - f_y_const

        from lenstronomy.LightModel.light_model import LightModel

        gauss = LightModel(["GAUSSIAN"])
        kwargs_source = [{"amp": 1, "sigma": 0.05, "center_x": 0, "center_y": 0}]
        flux_sis = gauss.surface_brightness(beta_x_sis, beta_y_sis, kwargs_source)
        flux_const = gauss.surface_brightness(beta_x_const, beta_y_const, kwargs_source)

        npt.assert_almost_equal(
            (flux_const - flux_sis) / np.max(flux_const), 0, decimal=2
        )

        # check for stability outside the defined bounds of curvature
        f_x_const, f_y_const = self.arc_const.derivatives(x=0, y=1000, **kwargs_arc)
        npt.assert_almost_equal(f_x_const, 0)
        npt.assert_almost_equal(f_y_const, 0)

    def test_hessian(self):
        kwargs_arc = {
            "tangential_stretch": 5,
            #'radial_stretch': 1,
            "curvature": 1.0 / 10,
            "direction": 0.5,
            "center_x": 0,
            "center_y": 0,
        }
        x, y = 0.0, 1.0
        f_xx, f_xy, f_yx, f_yy = self.arc_const.hessian(x, y, **kwargs_arc)

        alpha_ra, alpha_dec = self.arc_const.derivatives(x, y, **kwargs_arc)
        diff = 0.0000001
        alpha_ra_dx, alpha_dec_dx = self.arc_const.derivatives(
            x + diff, y, **kwargs_arc
        )
        alpha_ra_dy, alpha_dec_dy = self.arc_const.derivatives(
            x, y + diff, **kwargs_arc
        )

        f_xx_num = (alpha_ra_dx - alpha_ra) / diff
        f_xy_num = (alpha_ra_dy - alpha_ra) / diff
        f_yx_num = (alpha_dec_dx - alpha_dec) / diff
        f_yy_num = (alpha_dec_dy - alpha_dec) / diff
        print(f_xx, f_xx_num)
        print(f_xy, f_xy_num)
        print(f_yx, f_yx_num)
        print(f_yy, f_yy_num)

        npt.assert_almost_equal(f_xx, f_xx_num)
        npt.assert_almost_equal(f_xy, f_xy_num)
        npt.assert_almost_equal(f_yx, f_yx_num)
        npt.assert_almost_equal(f_yy, f_yy_num)


class TestCurvedArcConstMST(object):
    def setup_method(self):
        self.arc_const = CurvedArcConstMST()

    def test_function(self):
        kwargs_arc = {
            "tangential_stretch": 5,
            "radial_stretch": 1,
            "curvature": 1.0 / 10,
            "direction": 0,
            "center_x": 0,
            "center_y": 0,
        }
        npt.assert_raises(Exception, self.arc_const.function, 0.0, 0.0, **kwargs_arc)

    def test_hessian(self):
        kwargs_arc = {
            "tangential_stretch": 5,
            "radial_stretch": 1,
            "curvature": 1.0 / 10,
            "direction": 0,
            "center_x": 0,
            "center_y": 0,
        }
        # npt.assert_raises(Exception, self.arc_const.hessian, 0., 0., **kwargs_arc)


if __name__ == "__main__":
    pytest.main("-k TestLensModel")



from lenstronomy.SimulationAPI.model_api import ModelAPI
import pytest
import numpy.testing as npt
from lenstronomy.Cosmo.lens_cosmo import LensCosmo


class TestModelAPI(object):
    def setup_method(self):
        self.api = ModelAPI(
            lens_model_list=["SIS"],
            z_lens=None,
            z_source=None,
            lens_redshift_list=None,
            source_light_model_list=["GAUSSIAN"],
            lens_light_model_list=["SERSIC"],
            point_source_model_list=["UNLENSED"],
            source_redshift_list=None,
            cosmo=None,
        )

    def test_lens_model_class(self):
        model = self.api.lens_model_class
        assert model.lens_model_list[0] == "SIS"

    def test_lens_light_model_class(self):
        model = self.api.lens_light_model_class
        assert model.profile_type_list[0] == "SERSIC"

    def test_source_model_class(self):
        model = self.api.source_model_class
        assert model.profile_type_list[0] == "GAUSSIAN"

    def test_point_source_model_class(self):
        model = self.api.point_source_model_class
        assert model.point_source_type_list[0] == "UNLENSED"

    def test_source_position(self):
        api = ModelAPI(
            lens_model_list=["SIS"],
            z_lens=None,
            z_source=None,
            lens_redshift_list=None,
            source_light_model_list=["GAUSSIAN"],
            lens_light_model_list=["SERSIC"],
            point_source_model_list=["SOURCE_POSITION"],
            source_redshift_list=None,
            cosmo=None,
        )
        model = api.point_source_model_class
        assert model.point_source_type_list[0] == "SOURCE_POSITION"

    def test_physical2lensing_conversion(self):
        lens_redshift_list = [0.5, 1]
        z_source_convention = 2
        api = ModelAPI(
            lens_model_list=["SIS", "NFW"],
            lens_redshift_list=lens_redshift_list,
            z_source_convention=z_source_convention,
            cosmo=None,
            z_source=z_source_convention,
        )

        kwargs_mass = [
            {"sigma_v": 200, "center_x": 0, "center_y": 0},
            {"M200": 10**13, "concentration": 5, "center_x": 1, "center_y": 1},
        ]
        kwargs_lens = api.physical2lensing_conversion(kwargs_mass)

        theta_E = kwargs_lens[0]["theta_E"]
        lens_cosmo = LensCosmo(
            z_lens=lens_redshift_list[0], z_source=z_source_convention
        )
        theta_E_test = lens_cosmo.sis_sigma_v2theta_E(kwargs_mass[0]["sigma_v"])
        npt.assert_almost_equal(theta_E, theta_E_test, decimal=7)

        alpha_Rs = kwargs_lens[1]["alpha_Rs"]
        lens_cosmo = LensCosmo(
            z_lens=lens_redshift_list[1], z_source=z_source_convention
        )
        Rs_new, alpha_Rs_new = lens_cosmo.nfw_physical2angle(
            kwargs_mass[1]["M200"], kwargs_mass[1]["concentration"]
        )
        npt.assert_almost_equal(alpha_Rs, alpha_Rs_new, decimal=7)

        api = ModelAPI(
            lens_model_list=["SIS", "NFW"],
            z_lens=0.5,
            z_source_convention=z_source_convention,
            cosmo=None,
            z_source=z_source_convention,
        )
        kwargs_lens = api.physical2lensing_conversion(kwargs_mass)

        theta_E = kwargs_lens[0]["theta_E"]
        lens_cosmo = LensCosmo(z_lens=0.5, z_source=z_source_convention)
        theta_E_test = lens_cosmo.sis_sigma_v2theta_E(kwargs_mass[0]["sigma_v"])
        npt.assert_almost_equal(theta_E, theta_E_test, decimal=7)


if __name__ == "__main__":
    pytest.main()



__author__ = "gipagano"

import numpy as np
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["ConstMag"]


class ConstMag(LensProfileBase):
    """This class implements the macromodel potential of Diego et al.

    <https://www.aanda.org/articles/aa/pdf/2019/07/aa35490-19.pdf>`_
    Convergence and shear are computed according to `Diego2018 <arXiv:1706.10281v2>`_
    """

    param_names = ["center_x", "center_y", "mu_r", "mu_t", "parity", "phi_G"]
    lower_limit_default = {
        "center_x": -100,
        "center_y": -100,
        "mu_r": 1,
        "mu_t": 1000,
        "parity": -1,
        "phi_G": 0.0,
    }
    upper_limit_default = {
        "center_x": 100,
        "center_y": 100,
        "mu_r": 1,
        "mu_t": 1000,
        "parity": 1,
        "phi_G": np.pi,
    }

    def function(self, x, y, mu_r, mu_t, parity, phi_G, center_x=0, center_y=0):
        """

        :param x: x-coord (in angles)
        :param y: y-coord (in angles)
        :param mu_r: radial magnification
        :param mu_t: tangential magnification
        :param parity: parity side of the macromodel. Either +1 (positive parity) or -1 (negative parity)
        :param phi_G: shear orientation angle (relative to the x-axis)
        :return: lensing potential
        """

        # positive parity case
        if parity == 1:
            gamma = (1.0 / mu_t - 1.0 / mu_r) * 0.5
            kappa = 1 - gamma - 1.0 / mu_r

        # negative parity case
        elif parity == -1:
            gamma = (1.0 / mu_t + 1.0 / mu_r) * 0.5
            kappa = 1 - gamma + 1.0 / mu_r
        else:
            raise ValueError(
                "%f is not a valid value for the parity of the macromodel. Choose either +1 or -1."
                % parity
            )

        # compute the shear along the x and y directions, rotate the vector in the opposite direction than the reference frame (compare with util.rotate)
        gamma1, gamma2 = gamma * np.cos(2 * phi_G), -gamma * np.sin(2 * phi_G)

        x_shift = x - center_x
        y_shift = y - center_y
        f_ = (
            1.0 / 2.0 * kappa * (x_shift * x_shift + y_shift * y_shift)
            + 1.0 / 2.0 * gamma1 * (x_shift * x_shift - y_shift * y_shift)
            - gamma2 * x_shift * y_shift
        )

        return f_

    def derivatives(self, x, y, mu_r, mu_t, parity, phi_G, center_x=0, center_y=0):
        """

        :param x: x-coord (in angles)
        :param y: y-coord (in angles)
        :param mu_r: radial magnification
        :param mu_t: tangential magnification
        :param parity: parity of the side of the macromodel. Either +1 (positive parity) or -1 (negative parity)
        :param phi_G: shear orientation angle (relative to the x-axis)
        :return: deflection angle (in angles)
        """

        # positive parity case
        if parity == 1:
            gamma = (1.0 / mu_t - 1.0 / mu_r) * 0.5
            kappa = 1 - gamma - 1.0 / mu_r

        # negative parity case
        elif parity == -1:
            gamma = (1.0 / mu_t + 1.0 / mu_r) * 0.5
            kappa = 1 - gamma + 1.0 / mu_r
        else:
            raise ValueError(
                "%f is not a valid value for the parity of the macromodel. Choose either +1 or -1."
                % parity
            )

        # compute the shear along the x and y directions, rotate the vector in the opposite direction than the reference frame (compare with util.rotate)
        gamma1, gamma2 = gamma * np.cos(2 * phi_G), -gamma * np.sin(2 * phi_G)

        x_shift = x - center_x
        y_shift = y - center_y
        f_x = (kappa + gamma1) * x_shift - gamma2 * y_shift
        f_y = (kappa - gamma1) * y_shift - gamma2 * x_shift

        return f_x, f_y

    def hessian(self, x, y, mu_r, mu_t, parity, phi_G, center_x=0, center_y=0):
        """

        :param x: x-coord (in angles)
        :param y: y-coord (in angles)
        :param mu_r: radial magnification
        :param mu_t: tangential magnification
        :param parity: parity of the side of the macromodel. Either +1 (positive parity) or -1 (negative parity)
        :param phi_G: shear orientation angle (relative to the x-axis)
        :return: hessian matrix (in angles)
        """

        # positive parity case
        if parity == 1:
            gamma = (1.0 / mu_t - 1.0 / mu_r) * 0.5
            kappa = 1 - gamma - 1.0 / mu_r

        # negative parity case
        elif parity == -1:
            gamma = (1.0 / mu_t + 1.0 / mu_r) * 0.5
            kappa = 1 - gamma + 1.0 / mu_r
        else:
            raise ValueError(
                "%f is not a valid value for the parity of the macromodel. Choose either +1 or -1."
                % parity
            )

        # compute the shear along the x and y directions, rotate the vector in the opposite direction than the reference frame (compare with util.rotate)
        gamma1, gamma2 = gamma * np.cos(2 * phi_G), -gamma * np.sin(2 * phi_G)

        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = -gamma2

        return f_xx, f_xy, f_xy, f_yy



__author__ = "sibirrer"


from lenstronomy.LensModel.Profiles.p_jaffe_ellipse import PJaffe_Ellipse
import lenstronomy.Util.param_util as param_util

import numpy as np
import numpy.testing as npt
import pytest


class TestP_JAFFW(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.profile = PJaffe_Ellipse()

    def test_function(self):
        x = np.array([1])
        y = np.array([2])
        sigma0 = 1.0
        Ra, Rs = 0.5, 0.8
        q, phi_G = 0.8, 0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        values = self.profile.function(
            x, y, sigma0, Ra, Rs, e1, e2, center_x=0, center_y=0
        )
        npt.assert_almost_equal(values[0], 0.9091040398607811, decimal=8)
        x = np.array([0])
        y = np.array([0])

        values = self.profile.function(
            x, y, sigma0, Ra, Rs, e1, e2, center_x=0, center_y=0
        )
        npt.assert_almost_equal(values[0], 0.20267440905756931, decimal=8)

        x = np.array([2, 3, 4])
        y = np.array([1, 1, 1])
        values = self.profile.function(
            x, y, sigma0, Ra, Rs, e1, e2, center_x=0, center_y=0
        )
        npt.assert_almost_equal(values[0], 0.8327830942970774, decimal=8)
        npt.assert_almost_equal(values[1], 1.0233085474140422, decimal=8)
        npt.assert_almost_equal(values[2], 1.1868752663038047, decimal=8)

    def test_derivatives(self):
        x = np.array([1])
        y = np.array([2])
        sigma0 = 1.0
        Ra, Rs = 0.5, 0.8
        q, phi_G = 0.8, 0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_x, f_y = self.profile.derivatives(
            x, y, sigma0, Ra, Rs, e1, e2, center_x=0, center_y=0
        )
        npt.assert_almost_equal(f_x[0], 0.08130928181117723, decimal=8)
        npt.assert_almost_equal(f_y[0], 0.25409150565992883, decimal=8)
        x = np.array([0])
        y = np.array([0])
        f_x, f_y = self.profile.derivatives(
            x, y, sigma0, Ra, Rs, e1, e2, center_x=0, center_y=0
        )
        assert f_x[0] == 0
        assert f_y[0] == 0

        x = np.array([1, 3, 4])
        y = np.array([2, 1, 1])
        values = self.profile.derivatives(
            x, y, sigma0, Ra, Rs, e1, e2, center_x=0, center_y=0
        )
        npt.assert_almost_equal(values[0][0], 0.08130928181117723, decimal=8)
        npt.assert_almost_equal(values[1][0], 0.25409150565992883, decimal=8)
        npt.assert_almost_equal(values[0][1], 0.17711143165920576, decimal=8)
        npt.assert_almost_equal(values[1][1], 0.09224553732250299, decimal=8)

    def test_hessian(self):
        x = np.array([1])
        y = np.array([2])
        sigma0 = 1.0
        Ra, Rs = 0.5, 0.8
        q, phi_G = 0.8, 0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_xx, f_xy, f_yx, f_yy = self.profile.hessian(
            x, y, sigma0, Ra, Rs, e1, e2, center_x=0, center_y=0
        )
        npt.assert_almost_equal(f_xx[0], 0.06259391932550429, decimal=8)
        npt.assert_almost_equal(f_yy[0], -0.05572123112917993, decimal=8)
        npt.assert_almost_equal(f_xy[0], -0.058485405643460275, decimal=8)
        npt.assert_almost_equal(f_xy, f_yx, decimal=6)
        x = np.array([1, 3, 4])
        y = np.array([2, 1, 1])
        values = self.profile.hessian(
            x, y, sigma0, Ra, Rs, e1, e2, center_x=0, center_y=0
        )
        npt.assert_almost_equal(values[0][0], 0.06259391932550429, decimal=8)
        npt.assert_almost_equal(values[3][0], -0.05572123112917993, decimal=8)

    def test_mass_3d_lens(self):
        mass = self.profile.mass_3d_lens(r=1, sigma0=1, Ra=0.5, Rs=0.8, e1=0, e2=0)
        npt.assert_almost_equal(mass, 0.87077306005349242, decimal=8)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
from lenstronomy.Util import util
from lenstronomy.LensModel.Profiles.nfw import NFW
from lenstronomy.LensModel.Profiles.nfw_ellipse import NFW_ELLIPSE
from lenstronomy.LensModel.Profiles.cored_steep_ellipsoid import CSEProductAvgSet
import lenstronomy.Util.param_util as param_util

__all__ = ["NFW_ELLIPSE_CSE"]


class NFW_ELLIPSE_CSE(NFW_ELLIPSE):
    """
    this class contains functions concerning the NFW profile with an ellipticity defined in the convergence
    parameterization of alpha_Rs and Rs is the same as for the spherical NFW profile
    Approximation with CSE profile introduced by Oguri 2021: https://arxiv.org/pdf/2106.11464.pdf
    Match to NFW using CSEs is approximate: kappa matches to ~1-2%

    relation are: R_200 = c * Rs


    """

    profile_name = "NFW_ELLIPSE_CSE"
    param_names = ["Rs", "alpha_Rs", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "Rs": 0,
        "alpha_Rs": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "Rs": 100,
        "alpha_Rs": 10,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self, high_accuracy=True):
        """

        :param high_accuracy: if True uses a more accurate larger set of CSE profiles (see Oguri 2021)
        :type high_accuracy: boolean
        """
        self.cse_major_axis_set = CSEProductAvgSet()
        self.nfw = NFW()
        if high_accuracy is True:
            # Table 1 in Oguri 2021
            self._s_list = [
                1.082411e-06,
                8.786566e-06,
                3.292868e-06,
                1.860019e-05,
                3.274231e-05,
                6.232485e-05,
                9.256333e-05,
                1.546762e-04,
                2.097321e-04,
                3.391140e-04,
                5.178790e-04,
                8.636736e-04,
                1.405152e-03,
                2.193855e-03,
                3.179572e-03,
                4.970987e-03,
                7.631970e-03,
                1.119413e-02,
                1.827267e-02,
                2.945251e-02,
                4.562723e-02,
                6.782509e-02,
                1.596987e-01,
                1.127751e-01,
                2.169469e-01,
                3.423835e-01,
                5.194527e-01,
                8.623185e-01,
                1.382737e00,
                2.034929e00,
                3.402979e00,
                5.594276e00,
                8.052345e00,
                1.349045e01,
                2.603825e01,
                4.736823e01,
                6.559320e01,
                1.087932e02,
                1.477673e02,
                2.495341e02,
                4.305999e02,
                7.760206e02,
                2.143057e03,
                1.935749e03,
            ]
            self._a_list = [
                1.648988e-18,
                6.274458e-16,
                3.646620e-17,
                3.459206e-15,
                2.457389e-14,
                1.059319e-13,
                4.211597e-13,
                1.142832e-12,
                4.391215e-12,
                1.556500e-11,
                6.951271e-11,
                3.147466e-10,
                1.379109e-09,
                3.829778e-09,
                1.384858e-08,
                5.370951e-08,
                1.804384e-07,
                5.788608e-07,
                3.205256e-06,
                1.102422e-05,
                4.093971e-05,
                1.282206e-04,
                4.575541e-04,
                7.995270e-04,
                5.013701e-03,
                1.403508e-02,
                5.230727e-02,
                1.898907e-01,
                3.643448e-01,
                7.203734e-01,
                1.717667e00,
                2.217566e00,
                3.187447e00,
                8.194898e00,
                1.765210e01,
                1.974319e01,
                2.783688e01,
                4.482311e01,
                5.598897e01,
                1.426485e02,
                2.279833e02,
                5.401335e02,
                9.743682e02,
                1.775124e03,
            ]

        else:
            # Table 3 in Oguri 2021
            self._a_list = [
                1.434960e-16,
                5.232413e-14,
                2.666660e-12,
                7.961761e-11,
                2.306895e-09,
                6.742968e-08,
                1.991691e-06,
                5.904388e-05,
                1.693069e-03,
                4.039850e-02,
                5.665072e-01,
                3.683242e00,
                1.582481e01,
                6.340984e01,
                2.576763e02,
                1.422619e03,
            ]
            self._s_list = [
                4.041628e-06,
                3.086267e-05,
                1.298542e-04,
                4.131977e-04,
                1.271373e-03,
                3.912641e-03,
                1.208331e-02,
                3.740521e-02,
                1.153247e-01,
                3.472038e-01,
                1.017550e00,
                3.253031e00,
                1.190315e01,
                4.627701e01,
                1.842613e02,
                8.206569e02,
            ]

        super(NFW_ELLIPSE_CSE, self).__init__()

    def function(self, x, y, Rs, alpha_Rs, e1, e2, center_x=0, center_y=0):
        """Returns elliptically distorted NFW lensing potential.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param alpha_Rs: deflection (angular units) at projected Rs
        :param e1: eccentricity component in x-direction
        :param e2: eccentricity component in y-direction
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: lensing potential
        """
        phi_q, q = param_util.ellipticity2phi_q(e1, e2)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_q)

        # potential calculation
        f_ = self.cse_major_axis_set.function(
            x__ / Rs, y__ / Rs, self._a_list, self._s_list, q
        )
        const = self._normalization(alpha_Rs, Rs, q)
        return const * f_

    def derivatives(self, x, y, Rs, alpha_Rs, e1, e2, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function, calculated as an elliptically
        distorted deflection angle of the spherical NFW profile.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param alpha_Rs: deflection (angular units) at projected Rs
        :param e1: eccentricity component in x-direction
        :param e2: eccentricity component in y-direction
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: deflection in x-direction, deflection in y-direction
        """
        phi_q, q = param_util.ellipticity2phi_q(e1, e2)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_q)
        f__x, f__y = self.cse_major_axis_set.derivatives(
            x__ / Rs, y__ / Rs, self._a_list, self._s_list, q
        )

        # rotate deflections back
        f_x, f_y = util.rotate(f__x, f__y, -phi_q)
        const = self._normalization(alpha_Rs, Rs, q) / Rs
        return const * f_x, const * f_y

    def hessian(self, x, y, Rs, alpha_Rs, e1, e2, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^f/dy^2, d^2/dxdy the
        calculation is performed as a numerical differential from the deflection field.
        Analytical relations are possible.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param alpha_Rs: deflection (angular units) at projected Rs
        :param e1: eccentricity component in x-direction
        :param e2: eccentricity component in y-direction
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: d^2f/dx^2, d^2/dxdy, d^2/dydx, d^f/dy^2
        """
        phi_q, q = param_util.ellipticity2phi_q(e1, e2)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_q)
        f__xx, f__xy, __, f__yy = self.cse_major_axis_set.hessian(
            x__ / Rs, y__ / Rs, self._a_list, self._s_list, q
        )

        # rotate back
        kappa = 1.0 / 2 * (f__xx + f__yy)
        gamma1__ = 1.0 / 2 * (f__xx - f__yy)
        gamma2__ = f__xy
        gamma1 = np.cos(2 * phi_q) * gamma1__ - np.sin(2 * phi_q) * gamma2__
        gamma2 = +np.sin(2 * phi_q) * gamma1__ + np.cos(2 * phi_q) * gamma2__
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        const = self._normalization(alpha_Rs, Rs, q) / Rs**2

        return const * f_xx, const * f_xy, const * f_xy, const * f_yy

    def _normalization(self, alpha_Rs, Rs, q):
        """Applying to eqn 7 and 8 in Oguri 2021 from phenomenological definition.

        :param alpha_Rs: deflection at Rs
        :param Rs: scale radius
        :param q: axis ratio
        :return: normalization (m)
        """
        rho0 = self.nfw.alpha2rho0(alpha_Rs, Rs)
        rs_ = Rs
        const = 4 * rho0 * rs_**3
        return const



import numpy as np
import lenstronomy.Util.constants as const

__all__ = ["TimeDelayLikelihood"]


class TimeDelayLikelihood(object):
    """Class to compute the likelihood of a model given a measurement of time delays."""

    def __init__(
        self,
        time_delays_measured,
        time_delays_uncertainties,
        lens_model_class,
        point_source_class,
    ):
        """

        :param time_delays_measured: relative time delays (in days) in respect to the first image of the point source
        :param time_delays_uncertainties: time-delay uncertainties in same order as time_delay_measured. Alternatively
         a full covariance matrix that describes the likelihood.
        :param lens_model_class: instance of the LensModel() class
        :param point_source_class: instance of the PointSource() class, note: the first point source type is the one the
         time delays are imposed on
        """

        if time_delays_measured is None:
            raise ValueError(
                "time_delay_measured need to be specified to evaluate the time-delay likelihood."
            )
        if time_delays_uncertainties is None:
            raise ValueError(
                "time_delay_uncertainties need to be specified to evaluate the time-delay likelihood."
            )
        self._delays_measured = np.array(time_delays_measured)
        self._delays_errors = np.array(time_delays_uncertainties)
        self._lensModel = lens_model_class
        self._pointSource = point_source_class

    def logL(self, kwargs_lens, kwargs_ps, kwargs_cosmo):
        """Routine to compute the log likelihood of the time delay distance :param
        kwargs_lens: lens model kwargs list :param kwargs_ps: point source kwargs list
        :param kwargs_cosmo: cosmology and other kwargs :return: log likelihood of the
        model given the time delay data."""
        x_pos, y_pos = self._pointSource.image_position(
            kwargs_ps=kwargs_ps, kwargs_lens=kwargs_lens, original_position=True
        )
        x_pos, y_pos = x_pos[0], y_pos[0]
        delay_arcsec = self._lensModel.fermat_potential(x_pos, y_pos, kwargs_lens)
        D_dt_model = kwargs_cosmo["D_dt"]
        delay_days = const.delay_arcsec2days(delay_arcsec, D_dt_model)
        logL = self._logL_delays(delay_days, self._delays_measured, self._delays_errors)
        return logL

    @staticmethod
    def _logL_delays(delays_model, delays_measured, delays_errors):
        """Log likelihood of modeled delays vs measured time delays under considerations
        of errors.

        :param delays_model: n delays of the model (not relative delays)
        :param delays_measured: relative delays (1-2,1-3,1-4) relative to the first in
            the list
        :param delays_errors: gaussian errors on the measured delays
        :return: log likelihood of data given model
        """
        if len(delays_model) - 1 != len(delays_measured):
            return -(10**15)
        delta_t_model = np.array(delays_model[1:]) - delays_model[0]
        if delays_errors.ndim <= 1:
            logL = np.sum(
                -((delta_t_model - delays_measured) ** 2) / (2 * delays_errors**2)
            )
        elif delays_errors.ndim == 2:
            D = delta_t_model - delays_measured
            logL = (
                -1 / 2 * D @ np.linalg.inv(delays_errors) @ D
            )  # TODO: only calculate the inverse once
        else:
            raise ValueError(
                "Dimension of time delay error needs to be either one- or two-dimensional, not %s"
                % delays_errors.ndim
            )
        return logL

    @property
    def num_data(self):
        """

        :return: number of time delay measurements
        """
        return len(self._delays_measured)



import numpy as np
from lenstronomy.PointSource.Types.base_ps import (
    PSBase,
    _expand_to_array,
    _shrink_array,
)

__all__ = ["SourcePositions"]


class SourcePositions(PSBase):
    """Class of a single point source defined in the original source coordinate position
    that is lensed. The lens equation is solved to compute the image positions for the
    specified source position.

    Name within the PointSource module: 'SOURCE_POSITION'
    parameters: ra_source, dec_source, source_amp, mag_pert (optional)
    If fixed_magnification=True, than 'source_amp' is a parameter instead of 'point_amp'
    mag_pert is a list of fractional magnification pertubations applied to point source images
    """

    def image_position(
        self,
        kwargs_ps,
        kwargs_lens=None,
        magnification_limit=None,
        kwargs_lens_eqn_solver=None,
        **kwargs
    ):
        """On-sky image positions.

        :param kwargs_ps: keyword arguments of the point source model
        :param kwargs_lens: keyword argument list of the lens model(s), only used when
            requiring the lens equation solver
        :param magnification_limit: float >0 or None, if float is set and additional
            images are computed, only those images will be computed that exceed the
            lensing magnification (absolute value) limit
        :param kwargs_lens_eqn_solver: keyword arguments specifying the numerical
            settings for the lens equation solver see LensEquationSolver() class for
            details
        :return: image positions in x, y as arrays
        """
        if kwargs_lens_eqn_solver is None:
            kwargs_lens_eqn_solver = {}
        ra_source, dec_source = self.source_position(kwargs_ps)
        ra_image, dec_image = self._solver.image_position_from_source(
            ra_source,
            dec_source,
            kwargs_lens,
            magnification_limit=magnification_limit,
            **kwargs_lens_eqn_solver
        )
        return ra_image, dec_image

    def source_position(self, kwargs_ps, **kwargs):
        """Original source position (prior to lensing)

        :param kwargs_ps: point source keyword arguments
        :return: x, y position (as numpy arrays)
        """
        ra_source = kwargs_ps["ra_source"]
        dec_source = kwargs_ps["dec_source"]
        return np.array(ra_source), np.array(dec_source)

    def image_amplitude(
        self,
        kwargs_ps,
        kwargs_lens=None,
        x_pos=None,
        y_pos=None,
        magnification_limit=None,
        kwargs_lens_eqn_solver=None,
    ):
        """Image brightness amplitudes.

        :param kwargs_ps: keyword arguments of the point source model
        :param kwargs_lens: keyword argument list of the lens model(s), only ignored
            when providing image positions directly
        :param x_pos: pre-computed image position (no lens equation solver applied)
        :param y_pos: pre-computed image position (no lens equation solver applied)
        :param magnification_limit: float >0 or None, if float is set and additional
            images are computed, only those images will be computed that exceed the
            lensing magnification (absolute value) limit
        :param kwargs_lens_eqn_solver: keyword arguments specifying the numerical
            settings for the lens equation solver see LensEquationSolver() class for
            details
        :return: array of image amplitudes
        """
        if self._fixed_magnification:
            if x_pos is not None and y_pos is not None:
                ra_image, dec_image = x_pos, y_pos
            else:
                if kwargs_lens_eqn_solver is None:
                    kwargs_lens_eqn_solver = {}
                ra_image, dec_image = self.image_position(
                    kwargs_ps,
                    kwargs_lens=kwargs_lens,
                    magnification_limit=magnification_limit,
                    **kwargs_lens_eqn_solver
                )
            mag = self._lens_model.magnification(ra_image, dec_image, kwargs_lens)
            point_amp = kwargs_ps["source_amp"] * np.abs(mag)
        else:
            point_amp = kwargs_ps["point_amp"]
            if x_pos is not None:
                point_amp = _expand_to_array(point_amp, len(x_pos))
        mag_pert = kwargs_ps.get("mag_pert", 1)
        mag_pert = _shrink_array(mag_pert, len(point_amp))
        point_amp *= np.array(mag_pert)
        return np.array(point_amp)

    def source_amplitude(self, kwargs_ps, kwargs_lens=None):
        """Intrinsic brightness amplitude of point source When brightnesses are defined
        in magnified on-sky positions, the intrinsic brightness is computed as the mean
        in the magnification corrected image position brightnesses.

        :param kwargs_ps: keyword arguments of the point source model
        :param kwargs_lens: keyword argument list of the lens model(s), used when
            brightness are defined in magnified on-sky positions
        :return: brightness amplitude (as numpy array)
        """
        if self._fixed_magnification:
            source_amp = kwargs_ps["source_amp"]
        else:
            ra_image, dec_image = self.image_position(kwargs_ps, kwargs_lens)
            mag = self._lens_model.magnification(ra_image, dec_image, kwargs_lens)
            point_amp = kwargs_ps["point_amp"]
            source_amp = np.mean(np.array(point_amp) / np.array(mag))
        return np.array(source_amp)



__author__ = "sibirrer"


import numpy as np
import pytest
import numpy.testing as npt
import lenstronomy.Util.param_util as param_util
from lenstronomy.Util import util

try:
    import fastell4py

    fastell4py_bool = True
except:
    print(
        "Warning: fastell4py not available, tests will be trivially fulfilled without giving the right answer!"
    )
    fastell4py_bool = False


class TestEPLvsNIE(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        from lenstronomy.LensModel.Profiles.epl import EPL

        self.EPL = EPL()
        from lenstronomy.LensModel.Profiles.nie import NIE

        self.NIE = NIE()

    def test_function(self):
        phi_E = 1.0
        gamma = 2.0
        q = 0.999
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        x = np.array([1.0, 2])
        y = np.array([2, 0])
        values = self.EPL.function(x, y, phi_E, gamma, e1, e2)
        values_nie = self.NIE.function(x, y, phi_E, e1, e2, 0.0)
        delta_f = values[0] - values[1]
        delta_f_nie = values_nie[0] - values_nie[1]
        npt.assert_almost_equal(delta_f, delta_f_nie, decimal=5)

        q = 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        x = np.array([1.0, 2])
        y = np.array([2, 0])
        values = self.EPL.function(x, y, phi_E, gamma, e1, e2)
        values_nie = self.NIE.function(x, y, phi_E, e1, e2, 0.0)
        delta_f = values[0] - values[1]
        delta_f_nie = values_nie[0] - values_nie[1]
        npt.assert_almost_equal(delta_f, delta_f_nie, decimal=5)

        q = 0.4
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        x = np.array([1.0, 2])
        y = np.array([2, 0])
        values = self.EPL.function(x, y, phi_E, gamma, e1, e2)
        values_nie = self.NIE.function(x, y, phi_E, e1, e2, 0.0)
        delta_f = values[0] - values[1]
        delta_f_nie = values_nie[0] - values_nie[1]
        npt.assert_almost_equal(delta_f, delta_f_nie, decimal=5)

    def test_derivatives(self):
        x = np.array([1])
        y = np.array([2])
        phi_E = 1.0
        gamma = 2.0
        q = 1.0
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_x, f_y = self.EPL.derivatives(x, y, phi_E, gamma, e1, e2)
        f_x_nie, f_y_nie = self.NIE.derivatives(x, y, phi_E, e1, e2, 0.0)
        npt.assert_almost_equal(f_x, f_x_nie, decimal=4)
        npt.assert_almost_equal(f_y, f_y_nie, decimal=4)

        q = 0.7
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_x, f_y = self.EPL.derivatives(x, y, phi_E, gamma, e1, e2)
        f_x_nie, f_y_nie = self.NIE.derivatives(x, y, phi_E, e1, e2, 0.0)
        npt.assert_almost_equal(f_x, f_x_nie, decimal=4)
        npt.assert_almost_equal(f_y, f_y_nie, decimal=4)

    def test_hessian(self):
        x = np.array([1.0])
        y = np.array([2.0])
        phi_E = 1.0
        gamma = 2.0
        q = 0.9
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_xx, f_xy, f_yx, f_yy = self.EPL.hessian(x, y, phi_E, gamma, e1, e2)
        f_xx_nie, f_xy_nie, f_yx_nie, f_yy_nie = self.NIE.hessian(
            x, y, phi_E, e1, e2, 0.0
        )
        npt.assert_almost_equal(f_xx, f_xx_nie, decimal=4)
        npt.assert_almost_equal(f_yy, f_yy_nie, decimal=4)
        npt.assert_almost_equal(f_xy, f_xy_nie, decimal=4)
        npt.assert_almost_equal(f_xy, f_yx, decimal=8)

    def test_density_lens(self):
        r = 1
        kwargs = {"theta_E": 1, "gamma": 2, "e1": 0, "e2": 0}
        rho = self.EPL.density_lens(r, **kwargs)
        from lenstronomy.LensModel.Profiles.spep import SPEP

        spep = SPEP()
        rho_spep = spep.density_lens(r, **kwargs)
        npt.assert_almost_equal(rho, rho_spep, decimal=7)

    def test_mass_3d_lens(self):
        r = 1
        kwargs = {"theta_E": 1, "gamma": 2, "e1": 0, "e2": 0}
        rho = self.EPL.mass_3d_lens(r, **kwargs)
        from lenstronomy.LensModel.Profiles.spep import SPEP

        spep = SPEP()
        rho_spep = spep.mass_3d_lens(r, **kwargs)
        npt.assert_almost_equal(rho, rho_spep, decimal=7)

    def test_static(self):
        x, y = 1.0, 1.0
        phi_G, q = 0.3, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = {"theta_E": 1.0, "gamma": 1.5, "e1": e1, "e2": e2}
        f_ = self.EPL.function(x, y, **kwargs_lens)
        self.EPL.set_static(**kwargs_lens)
        f_static = self.EPL.function(x, y, **kwargs_lens)
        npt.assert_almost_equal(f_, f_static, decimal=8)
        self.EPL.set_dynamic()
        kwargs_lens = {"theta_E": 2.0, "gamma": 1.9, "e1": e1, "e2": e2}
        f_dyn = self.EPL.function(x, y, **kwargs_lens)
        assert f_dyn != f_static

    def test_regularization(self):
        phi_E = 1.0
        gamma = 2.0
        q = 1.0
        phi_G = 1.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)

        x = 0.0
        y = 0.0
        f_x, f_y = self.EPL.derivatives(x, y, phi_E, gamma, e1, e2)
        npt.assert_almost_equal(f_x, 0.0)
        npt.assert_almost_equal(f_y, 0.0)

        x = 0.0
        y = 0.0
        f_xx, f_xy, f_yx, f_yy = self.EPL.hessian(x, y, phi_E, gamma, e1, e2)
        assert f_xx > 10**5
        assert f_yy > 10**5
        # npt.assert_almost_equal(f_xx, 10**10)
        # npt.assert_almost_equal(f_yy, 10**10)
        npt.assert_almost_equal(f_xy, 0)
        npt.assert_almost_equal(f_yx, 0)


class TestEPLvsPEMD(object):
    """Test EPL model vs PEMD with FASTELL This tests get only executed if fastell is
    installed."""

    def setup_method(self):
        try:
            import fastell4py

            self._fastell4py_bool = True
        except:
            print(
                "Warning: fastell4py not available, tests will be trivially fulfilled without giving the right answer!"
            )
            self._fastell4py_bool = False
        from lenstronomy.LensModel.Profiles.epl import EPL

        self.epl = EPL()
        from lenstronomy.LensModel.Profiles.pemd import PEMD

        self.pemd = PEMD(suppress_fastell=True)

    def test_epl_pemd_convention(self):
        """Tests convention of EPL and PEMD model on the deflection angle basis."""
        if self._fastell4py_bool is False:
            assert True
        else:
            x, y = util.make_grid(numPix=10, deltapix=0.2)

            theta_E_list = [0.5, 1, 2]
            gamma_list = [1.8, 2.0, 2.2]
            e1_list = [-0.2, 0.0, 0.2]
            e2_list = [-0.2, 0.0, 0.2]
            for gamma in gamma_list:
                for e1 in e1_list:
                    for e2 in e2_list:
                        for theta_E in theta_E_list:
                            kwargs = {
                                "theta_E": theta_E,
                                "gamma": gamma,
                                "e1": e1,
                                "e2": e2,
                                "center_x": 0,
                                "center_y": 0,
                            }
                            f_x, f_y = self.epl.derivatives(x, y, **kwargs)
                            f_x_pemd, f_y_pemd = self.pemd.derivatives(x, y, **kwargs)

                            npt.assert_almost_equal(f_x, f_x_pemd, decimal=4)
                            npt.assert_almost_equal(f_y, f_y_pemd, decimal=4)


class TestEPLQPhi(object):
    """Tests if EPL and EPL_Q_PHI return same values."""

    def setup_method(self):
        from lenstronomy.LensModel.Profiles.epl import EPL

        self.epl = EPL()
        from lenstronomy.LensModel.Profiles.epl import EPLQPhi

        self.epl_qphi = EPLQPhi()

    def test_epl_eplqphi(self):
        x, y = util.make_grid(numPix=10, deltapix=0.2)

        theta_E_list = [0.5, 1, 2]
        gamma_list = [1.8, 2.0, 2.2]
        e1_list = [-0.2, 0.0, 0.2]
        e2_list = [-0.2, 0.0, 0.2]
        for gamma in gamma_list:
            for e1 in e1_list:
                for e2 in e2_list:
                    phi, q = param_util.ellipticity2phi_q(e1, e2)
                    for theta_E in theta_E_list:
                        kwargs = {
                            "theta_E": theta_E,
                            "gamma": gamma,
                            "e1": e1,
                            "e2": e2,
                            "center_x": 0,
                            "center_y": 0,
                        }
                        f_x, f_y = self.epl.derivatives(x, y, **kwargs)
                        kwargs_qphi = {
                            "theta_E": theta_E,
                            "gamma": gamma,
                            "q": q,
                            "phi": phi,
                            "center_x": 0,
                            "center_y": 0,
                        }
                        f_x_qphi, f_y_qphi = self.epl_qphi.derivatives(
                            x, y, **kwargs_qphi
                        )

                        npt.assert_almost_equal(f_x, f_x_qphi, decimal=4)
                        npt.assert_almost_equal(f_y, f_y_qphi, decimal=4)


if __name__ == "__main__":
    pytest.main()



from lenstronomy.LightModel.light_model import LightModel
import numpy as np

__all__ = ["DifferentialExtinction"]


class DifferentialExtinction(object):
    """Class to compute an extinction (for a specific band/wavelength).

    This class uses the functionality available in the LightModel module to describe an
    optical depth tau_ext to compute the extinction on the sky/image.
    """

    def __init__(self, optical_depth_model=None, tau0_index=0):
        """

        :param optical_depth_model: list of strings naming the profiles (same convention as LightModel module)
         describing the optical depth of the extinction
        """
        if optical_depth_model is None:
            optical_depth_model = []
        self._profile = LightModel(light_model_list=optical_depth_model)
        if len(optical_depth_model) == 0:
            self._compute_bool = False
        else:
            self._compute_bool = True
        self._tau0_index = tau0_index

    @property
    def compute_bool(self):
        """
        :return: True when a differential extinction is set, False otherwise
        """
        return self._compute_bool

    def extinction(self, x, y, kwargs_extinction=None, kwargs_special=None):
        """

        :param x: coordinate in image plane of flux intensity
        :param y: coordinate in image plane of flux intensity
        :param kwargs_extinction: keyword argument list matching the extinction profile
        :param kwargs_special: keyword arguments hosting special parameters, here required 'tau0_list'
        :return: extinction corrected flux
        """
        if self._compute_bool is False or kwargs_extinction is None:
            return 1
        tau = self._profile.surface_brightness(x, y, kwargs_list=kwargs_extinction)
        tau0_list = kwargs_special.get("tau0_list", None)
        if tau0_list is not None:
            tau0 = tau0_list[self._tau0_index]
        else:
            tau0 = 1
        return np.exp(-tau0 * tau)



__author__ = "sibirrer"

import numpy.testing as npt
import pytest
import numpy as np

from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF
from lenstronomy.ImSim.MultiBand.multi_linear import MultiLinear
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.ImSim.image_model import ImageModel
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver


class TestImageModel(object):
    """Tests the source model routines."""

    def setup_method(self):
        # data specifics
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 100  # cutout pixel size
        deltaPix = 0.05  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        # PSF specification

        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        data_class = ImageData(**kwargs_data)
        kwargs_psf = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "truncation": 5,
            "pixel_size": deltaPix,
        }
        psf_class = PSF(**kwargs_psf)
        # 'EXTERNAL_SHEAR': external shear
        kwargs_shear = {
            "gamma1": 0.01,
            "gamma2": 0.01,
        }  # gamma_ext: shear strength, psi_ext: shear angel (in radian)
        phi, q = 0.2, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": 1.8,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_model_list = ["SPEP", "SHEAR"]
        self.kwargs_lens = [kwargs_spemd, kwargs_shear]
        lens_model_class = LensModel(lens_model_list=lens_model_list)
        # list of light profiles (for lens and source)
        # 'SERSIC': spherical Sersic profile
        kwargs_sersic = {
            "amp": 1.0,
            "R_sersic": 0.1,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        # 'SERSIC_ELLIPSE': elliptical Sersic profile
        phi, q = 0.2, 0.9
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 7,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_light_model_list = ["SERSIC"]
        self.kwargs_lens_light = [kwargs_sersic]
        lens_light_model_class = LightModel(light_model_list=lens_light_model_list)
        source_model_list = ["SERSIC_ELLIPSE"]
        self.kwargs_source = [kwargs_sersic_ellipse]
        source_model_class = LightModel(light_model_list=source_model_list)
        self.kwargs_ps = [
            {"ra_source": 0.0001, "dec_source": 0.0, "source_amp": 1.0}
        ]  # quasar point source position in the source plane and intrinsic brightness
        point_source_class = PointSource(
            point_source_type_list=["SOURCE_POSITION"], fixed_magnification_list=[True]
        )
        kwargs_numerics = {"supersampling_factor": 2, "supersampling_convolution": True}
        imageModel = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            kwargs_numerics=kwargs_numerics,
        )
        image_sim = sim_util.simulate_simple(
            imageModel,
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
        )
        data_class.update_data(image_sim)
        kwargs_data["image_data"] = image_sim
        self.solver = LensEquationSolver(lensModel=lens_model_class)
        multi_band_list = [[kwargs_data, kwargs_psf, kwargs_numerics]]
        kwargs_model = {
            "lens_model_list": lens_model_list,
            "source_light_model_list": source_model_list,
            "point_source_model_list": ["SOURCE_POSITION"],
            "fixed_magnification_list": [True],
        }
        self.imageModel = MultiLinear(
            multi_band_list, kwargs_model, likelihood_mask_list=None, compute_bool=None
        )

    def test_image_linear_solve(self):
        model, error_map, cov_param, param = self.imageModel.image_linear_solve(
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
            inv_bool=False,
        )
        chi2_reduced = self.imageModel._imageModel_list[0].reduced_chi2(
            model[0], error_map[0]
        )
        npt.assert_almost_equal(chi2_reduced, 1, decimal=1)
        chi2_reduced_list = self.imageModel.reduced_residuals(
            model_list=model, error_map_list=error_map
        )
        npt.assert_almost_equal(
            np.sum(chi2_reduced_list[0] ** 2) / (100**2), 1, decimal=1
        )

    def test_likelihood_data_given_model(self):
        logL = self.imageModel.likelihood_data_given_model(
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
            source_marg=False,
        )
        npt.assert_almost_equal(logL, -5100, decimal=-3)

        logLmarg = self.imageModel.likelihood_data_given_model(
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
            source_marg=True,
        )
        npt.assert_almost_equal(logL - logLmarg, 0, decimal=-2)

    def test_numData_evaluate(self):
        numData = self.imageModel.num_data_evaluate
        assert numData == 10000


if __name__ == "__main__":
    pytest.main()



import lenstronomy.Util.util as util


instrument_name_list = ["LSST"]
observation_name_list = ["LSST_g_band", "LSST_r_band", "LSST_i_band"]

__all__ = ["observation_constructor"]


def observation_constructor(instrument_name, observation_name):
    """

    :param instrument_name: string, name of instrument referenced in this file
    :param observation_name: string, name of observation referenced in this file
    :return: instance of the SimulationAPI.data_type instance
    """

    if instrument_name == "LSST":
        kwargs_instrument = LSST_camera
    else:
        raise ValueError(
            "instrument name %s not supported! Choose among %s"
            % (instrument_name, instrument_name_list)
        )

    if observation_name == "LSST_g_band":
        kwargs_observation = LSST_g_band_obs
    elif observation_name == "LSST_r_band":
        kwargs_observation = LSST_r_band_obs
    elif observation_name == "LSST_i_band":
        kwargs_observation = LSST_i_band_obs
    else:
        raise ValueError(
            "observation name %s not supported! Choose among %s"
            % (observation_name, observation_name_list)
        )
    kwargs_data = util.merge_dicts(kwargs_instrument, kwargs_observation)
    return kwargs_data


LSST_camera = {"read_noise": 10, "pixel_scale": 0.263, "ccd_gain": 4.5}

LSST_g_band_obs = {
    "exposure_time": 90.0,
    "sky_brightness": 21.7,
    "magnitude_zero_point": 30,
    "num_exposures": 10,
    "seeing": 0.9,
    "psf_type": "GAUSSIAN",
}

LSST_r_band_obs = {
    "exposure_time": 90.0,
    "sky_brightness": 20.7,
    "magnitude_zero_point": 30,
    "num_exposures": 10,
    "seeing": 0.9,
    "psf_type": "GAUSSIAN",
}

LSST_i_band_obs = {
    "exposure_time": 90.0,
    "sky_brightness": 20.1,
    "magnitude_zero_point": 30,
    "num_exposures": 10,
    "seeing": 0.9,
    "psf_type": "GAUSSIAN",
}



__author__ = "sibirrer"


import lenstronomy.Util.class_creator as class_creator
import pytest
import numpy as np
import unittest


class TestClassCreator(object):
    def setup_method(self):
        self.kwargs_model = {
            "lens_model_list": ["SIS"],
            "source_light_model_list": ["SERSIC"],
            "lens_light_model_list": ["SERSIC"],
            "point_source_model_list": ["LENSED_POSITION"],
            "index_lens_model_list": [[0]],
            "index_source_light_model_list": [[0]],
            "index_lens_light_model_list": [[0]],
            "index_point_source_model_list": [[0]],
            "band_index": 0,
            "source_deflection_scaling_list": [1],
            "source_redshift_list": [1],
            "fixed_magnification_list": [True],
            "additional_images_list": [False],
            "lens_redshift_list": [0.5],
            "point_source_frame_list": [[0]],
        }
        self.kwargs_model_2 = {
            "lens_model_list": ["SIS"],
            "source_light_model_list": ["SERSIC"],
            "lens_light_model_list": ["SERSIC"],
            "point_source_model_list": ["LENSED_POSITION"],
        }
        self.kwargs_model_3 = {
            "lens_model_list": ["SIS"],
            "source_light_model_list": ["SERSIC"],
            "lens_light_model_list": ["SERSIC"],
            "point_source_model_list": ["LENSED_POSITION"],
            "index_lens_model_list": [[0]],
            "index_source_light_model_list": [[0]],
            "index_lens_light_model_list": [[0]],
            "index_point_source_model_list": [[0]],
            "point_source_frame_list": [[0]],
        }
        self.kwargs_model_4 = {
            "lens_model_list": ["SIS", "SIS"],
            "lens_redshift_list": [0.3, 0.4],
            "multi_plane": True,
            "observed_convention_index": [0],
            "index_lens_model_list": [[0]],
            "z_source": 1,
            "optical_depth_model_list": ["UNIFORM"],
            "index_optical_depth_model_list": [[0]],
            "tau0_index_list": [0],
            "point_source_frame_list": [[0]],
        }

        self.kwargs_psf = {"psf_type": "NONE"}
        self.kwargs_data = {"image_data": np.ones((10, 10))}

    def test_create_class_instances(self):
        (
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            extinction_class,
        ) = class_creator.create_class_instances(**self.kwargs_model)
        assert lens_model_class.lens_model_list[0] == "SIS"

        (
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            extinction_class,
        ) = class_creator.create_class_instances(**self.kwargs_model_2)
        assert lens_model_class.lens_model_list[0] == "SIS"

        (
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            extinction_class,
        ) = class_creator.create_class_instances(**self.kwargs_model_3)
        assert lens_model_class.lens_model_list[0] == "SIS"

        (
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            extinction_class,
        ) = class_creator.create_class_instances(**self.kwargs_model_4)
        assert lens_model_class.lens_model_list[0] == "SIS"
        assert lens_model_class.lens_model._observed_convention_index[0] == 0

    def test_create_image_model(self):
        imageModel = class_creator.create_image_model(
            self.kwargs_data,
            self.kwargs_psf,
            kwargs_numerics={},
            kwargs_model=self.kwargs_model,
        )
        assert imageModel.LensModel.lens_model_list[0] == "SIS"

        imageModel = class_creator.create_image_model(
            self.kwargs_data, self.kwargs_psf, kwargs_numerics={}, kwargs_model={}
        )
        assert imageModel.LensModel.lens_model_list == []

    def test_create_im_sim(self):
        kwargs_model = {
            "lens_model_list": ["SIS"],
            "source_light_model_list": ["SERSIC"],
            "lens_light_model_list": ["SERSIC"],
            "point_source_model_list": ["LENSED_POSITION"],
        }
        kwargs_psf = {"psf_type": "NONE"}
        kwargs_data = {"image_data": np.ones((10, 10))}

        multi_band_list = [[kwargs_data, kwargs_psf, {}]]
        multi_band_type = "multi-linear"

        multi_band = class_creator.create_im_sim(
            multi_band_list,
            multi_band_type,
            kwargs_model,
            bands_compute=None,
            image_likelihood_mask_list=None,
            band_index=0,
        )
        assert multi_band._imageModel_list[0].LensModel.lens_model_list[0] == "SIS"
        multi_band_type = "joint-linear"
        multi_band = class_creator.create_im_sim(
            multi_band_list,
            multi_band_type,
            kwargs_model,
            bands_compute=None,
            image_likelihood_mask_list=None,
            band_index=0,
        )
        assert multi_band._imageModel_list[0].LensModel.lens_model_list[0] == "SIS"
        multi_band_type = "single-band"
        multi_band = class_creator.create_im_sim(
            multi_band_list,
            multi_band_type,
            kwargs_model,
            bands_compute=None,
            image_likelihood_mask_list=None,
            band_index=0,
        )
        assert multi_band.LensModel.lens_model_list[0] == "SIS"


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            class_creator.create_im_sim(
                multi_band_list=None,
                multi_band_type="WRONG",
                kwargs_model=None,
                bands_compute=None,
                image_likelihood_mask_list=None,
                band_index=0,
            )
        with self.assertRaises(ValueError):
            class_creator.create_im_sim(
                multi_band_list=[[], []],
                multi_band_type="multi-linear",
                linear_solver=False,
                kwargs_model=None,
                bands_compute=None,
                image_likelihood_mask_list=None,
                band_index=0,
            )


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import pytest
import numpy as np
import numpy.testing as npt
import lenstronomy.Util.simulation_util as sim_util
import lenstronomy.Util.class_creator as class_creator
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.Sampling.likelihood import LikelihoodModule
from lenstronomy.Sampling.parameters import Param
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF
from lenstronomy.Data.kinematic_bin_2D import KinBin
from lenstronomy.Sampling.Likelihoods import kinematic_NN_call
import lenstronomy.Util.kernel_util as kernel_util


class TestLikelihoodModule(object):
    """Test the fitting sequences."""

    def setup_method(self):
        np.random.seed(42)

        # data specifics
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 50  # cutout pixel size
        deltaPix = 0.1  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        kwargs_model = {
            "lens_model_list": ["SPEP"],
            "lens_light_model_list": ["SERSIC"],
            "source_light_model_list": ["SERSIC"],
            "point_source_model_list": ["SOURCE_POSITION"],
            "fixed_magnification_list": [True],
        }

        # PSF specification
        kwargs_band = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        data_class = ImageData(**kwargs_band)
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": fwhm, "pixel_size": deltaPix}
        psf_class = PSF(**kwargs_psf)
        print(np.shape(psf_class.kernel_point_source), "test kernel shape -")
        kwargs_spep = {
            "theta_E": 1.0,
            "gamma": 1.95,
            "center_x": 0,
            "center_y": 0,
            "e1": 0.1,
            "e2": 0.1,
        }

        self.kwargs_lens = [kwargs_spep]
        kwargs_sersic = {
            "amp": 1 / 0.05**2.0,
            "R_sersic": 0.1,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        # 'SERSIC_ELLIPSE': elliptical Sersic profile
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 3,
            "center_x": 0,
            "center_y": 0,
        }

        self.kwargs_lens_light = [kwargs_sersic]
        self.kwargs_source = [kwargs_sersic_ellipse]
        self.kwargs_ps = [
            {"ra_source": 0.05, "dec_source": 0.02, "source_amp": 1.0}
        ]  # quasar point source position in the source plane and intrinsic brightness
        self.kwargs_cosmo = {"D_dt": 1000}
        kwargs_numerics = {
            "supersampling_factor": 1,
            "supersampling_convolution": False,
        }
        (
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            extinction_class,
        ) = class_creator.create_class_instances(**kwargs_model)
        imageModel = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            extinction_class,
            kwargs_numerics=kwargs_numerics,
        )
        image_sim = sim_util.simulate_simple(
            imageModel,
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
        )
        ra_pos, dec_pos = imageModel.PointSource.image_position(
            kwargs_ps=self.kwargs_ps, kwargs_lens=self.kwargs_lens
        )

        data_class.update_data(image_sim)
        kwargs_band["image_data"] = image_sim
        self.data_class = data_class
        self.psf_class = psf_class

        self.kwargs_model = kwargs_model
        self.kwargs_numerics = {
            "supersampling_factor": 1,
            "supersampling_convolution": False,
        }

        kwargs_constraints = {
            "num_point_source_list": [4],
            "solver_type": "NONE",  # 'PROFILE', 'PROFILE_SHEAR', 'ELLIPSE', 'CENTER'
            "Ddt_sampling": True,
        }

        def condition_definition(
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light,
            kwargs_ps=None,
            kwargs_special=None,
            kwargs_extinction=None,
        ):
            logL = 0
            if kwargs_lens_light[0]["R_sersic"] > kwargs_source[0]["R_sersic"]:
                logL -= 10**15
            return logL

        kwargs_likelihood = {
            "force_no_add_image": True,
            "source_marg": True,
            "astrometric_likelihood": True,
            "image_position_uncertainty": 0.004,
            "check_matched_source_position": False,
            "source_position_tolerance": 0.001,
            "source_position_sigma": 0.001,
            "check_positive_flux": True,
            "flux_ratio_likelihood": True,
            "prior_lens": [[0, "theta_E", 1, 0.1]],
            "custom_logL_addition": condition_definition,
            "image_position_likelihood": True,
        }
        self.kwargs_data = {
            "multi_band_list": [[kwargs_band, kwargs_psf, kwargs_numerics]],
            "multi_band_type": "single-band",
            "time_delays_measured": np.ones(3),
            "time_delays_uncertainties": np.ones(3),
            "flux_ratios": np.ones(3),
            "flux_ratio_errors": np.ones(3),
            "ra_image_list": ra_pos,
            "dec_image_list": dec_pos,
        }
        self.param_class = Param(self.kwargs_model, **kwargs_constraints)
        self.imageModel = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            kwargs_numerics=kwargs_numerics,
        )
        self.Likelihood = LikelihoodModule(
            kwargs_data_joint=self.kwargs_data,
            kwargs_model=kwargs_model,
            param_class=self.param_class,
            **kwargs_likelihood
        )
        self.kwargs_band = kwargs_band
        self.kwargs_psf = kwargs_psf
        self.numPix = numPix

    def test_logL(self):
        args = self.param_class.kwargs2args(
            kwargs_lens=self.kwargs_lens,
            kwargs_source=self.kwargs_source,
            kwargs_lens_light=self.kwargs_lens_light,
            kwargs_ps=self.kwargs_ps,
            kwargs_special=self.kwargs_cosmo,
        )

        logL = self.Likelihood.logL(args, verbose=True)
        num_data_evaluate = self.Likelihood.num_data
        npt.assert_almost_equal(logL / num_data_evaluate, -1 / 2.0, decimal=1)

    def test_time_delay_likelihood(self):
        kwargs_likelihood = {
            "time_delay_likelihood": True,
        }
        likelihood = LikelihoodModule(
            kwargs_data_joint=self.kwargs_data,
            kwargs_model=self.kwargs_model,
            param_class=self.param_class,
            **kwargs_likelihood
        )
        args = self.param_class.kwargs2args(
            kwargs_lens=self.kwargs_lens,
            kwargs_source=self.kwargs_source,
            kwargs_lens_light=self.kwargs_lens_light,
            kwargs_ps=self.kwargs_ps,
            kwargs_special=self.kwargs_cosmo,
        )

        logL = likelihood.logL(args, verbose=True)
        npt.assert_almost_equal(logL, -1328.821179288249, decimal=-1)

    def test_kin_likelihood(self):
        self.kinematic_NN = kinematic_NN_call.KinematicNN().SKiNN_installed
        if self.kinematic_NN:
            # since this doesn't work for the SPEP model, test it with a different lens model:
            kwargs_eplqphi = {
                "theta_E": 1.0,
                "gamma": 1.95,
                "center_x": 0,
                "center_y": 0,
                "q": 0.7,
                "phi": 0.7,
            }
            kwargs_lens = [kwargs_eplqphi]
            kwargs_sersic_ellipse_qphi = {
                "amp": 1.0,
                "R_sersic": 0.51,
                "n_sersic": 2.5,
                "center_x": 0,
                "center_y": 0,
                "q": 0.99,
                "phi": 0.7,
            }
            kwargs_lens_light = [kwargs_sersic_ellipse_qphi]
            kwargs_model = {
                "lens_model_list": ["EPL_Q_PHI"],
                "lens_light_model_list": ["SERSIC_ELLIPSE_Q_PHI"],
                "source_light_model_list": ["SERSIC"],
                "point_source_model_list": ["SOURCE_POSITION"],
                "fixed_magnification_list": [True],
                "z_lens": 1,
            }
            kwargs_constraints = {
                "num_point_source_list": [4],
                "solver_type": "NONE",  # 'PROFILE', 'PROFILE_SHEAR', 'ELLIPSE', 'CENTER'
                "Ddt_sampling": True,
            }
            kwargs_likelihood = {
                "kinematic_2d_likelihood": False,
                "kin_lens_idx": 0,
                "kin_lens_light_idx": 0,
                "time_delay_likelihood": False,
            }
            param_class = Param(kwargs_model, **kwargs_constraints)
            Likelihood = LikelihoodModule(
                kwargs_data_joint=self.kwargs_data,
                kwargs_model=kwargs_model,
                param_class=param_class,
                **kwargs_likelihood
            )
            kwargs_cosmo = {
                "D_dt": 2000,
                "D_d": 1000,
                "b_ani": 0.2,
                "incli": np.pi / 2 - 0.01,
            }
            args = param_class.kwargs2args(
                kwargs_lens=kwargs_lens,
                kwargs_source=self.kwargs_source,
                kwargs_lens_light=kwargs_lens_light,
                kwargs_ps=self.kwargs_ps,
                kwargs_special=kwargs_cosmo,
            )
            logL_nokin = Likelihood.logL(args, verbose=True)
            # logL should be same order as SPEP value, buthas different r_sersic so it will be worse
            npt.assert_allclose(logL_nokin, -1328.821179288249, rtol=1)

            # Now add kinematic likelihood
            # for simplicity, set kin image data to same as light data
            numPix = 50  # cutout pixel size
            deltaPix = 0.1  # pixel size in arcsec (area per pixel = deltaPix**2)

            binmap = np.zeros_like(
                self.kwargs_band["image_data"]
            )  # one single bin across whole image
            binned_dummy_data = np.array([200])
            delta_pix_kin = deltaPix
            npix_kin = numPix

            kwargs_kin = {
                "bin_data": binned_dummy_data,
                "bin_cov": np.diag((binned_dummy_data * 0.05) ** 2),  # 5% error
                "bin_mask": binmap,
                "ra_at_xy_0": -(npix_kin - 1) / 2.0 * delta_pix_kin,
                "dec_at_xy_0": -(npix_kin - 1) / 2.0 * delta_pix_kin,
                "transform_pix2angle": np.array([[1, 0], [0, 1]]) * delta_pix_kin,
            }

            kinkernel_point_source = kernel_util.kernel_gaussian(
                num_pix=9, delta_pix=0.2, fwhm=1.0
            )
            kwargs_pixelkin = {
                "psf_type": "PIXEL",
                "kernel_point_source": kinkernel_point_source,
            }
            kinPSF = PSF(**kwargs_pixelkin)
            _KinBin = KinBin(psf_class=kinPSF, **kwargs_kin)
            kwargs_data_kin = (
                self.kwargs_data.copy()
            )  # add kinematics data to kwargs_data
            kwargs_data_kin["kinematic_data"] = _KinBin

            # confirm that full likelihood is now lens+kin
            kwargs_likelihood = {
                "kinematic_2d_likelihood": True,
                "kin_lens_idx": 0,
                "kin_lens_light_idx": 0,
                "time_delay_likelihood": False,
            }

            kwargs_constraints["kinematic_sampling"] = True
            param_class = Param(kwargs_model, **kwargs_constraints)
            Likelihood = LikelihoodModule(
                kwargs_data_joint=kwargs_data_kin,
                kwargs_model=kwargs_model,
                param_class=param_class,
                **kwargs_likelihood
            )
            args = param_class.kwargs2args(
                kwargs_lens=kwargs_lens,
                kwargs_source=self.kwargs_source,
                kwargs_lens_light=kwargs_lens_light,
                kwargs_ps=self.kwargs_ps,
                kwargs_special=kwargs_cosmo,
            )

            logL = Likelihood.logL(args, verbose=True)
            # With only one bin, the new logL should be worse than without kin by
            # half the mean chi2 averaged over the whole image
            image_averaged_chi2 = (
                (np.mean(Likelihood.kinematic_2D_likelihood.vrms) - binned_dummy_data)
                ** 2
                / kwargs_kin["bin_cov"]
            ).flatten()
            npt.assert_almost_equal(logL_nokin - logL, image_averaged_chi2 / 2)

    def test_check_bounds(self):
        penalty, bound_hit = self.Likelihood.check_bounds(
            args=[0, 1], lowerLimit=[1, 0], upperLimit=[2, 2], verbose=True
        )
        assert bound_hit

    def test_pixelbased_modelling(self):
        ss_source = 2
        numPix_source = self.numPix * ss_source
        n_scales = 3
        kwargs_pixelbased = {
            "source_interpolation": "nearest",
            "supersampling_factor_source": ss_source,  # supersampling of pixelated source grid
            # following choices are to minimize pixel solver runtime (not to get accurate reconstruction!)
            "threshold_decrease_type": "none",
            "num_iter_source": 2,
            "num_iter_lens": 2,
            "num_iter_global": 2,
            "num_iter_weights": 2,
        }
        kwargs_likelihood = {
            "image_likelihood": True,
            "kwargs_pixelbased": kwargs_pixelbased,
            "check_positive_flux": True,  # effectively not applied, activated for code coverage purposes
        }
        kernel = PSF(**self.kwargs_psf).kernel_point_source
        kwargs_psf = {"psf_type": "PIXEL", "kernel_point_source": kernel}
        kwargs_numerics = {"supersampling_factor": 1}
        kwargs_data = {
            "multi_band_list": [[self.kwargs_band, kwargs_psf, kwargs_numerics]]
        }
        kwargs_model = {
            "lens_model_list": ["SPEP"],
            "lens_light_model_list": ["SLIT_STARLETS"],
            "source_light_model_list": ["SLIT_STARLETS"],
        }

        kwargs_fixed_source = [
            {
                "n_scales": n_scales,
                "n_pixels": numPix_source**2,
                "scale": 1,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        kwargs_fixed_lens_light = [
            {
                "n_scales": n_scales,
                "n_pixels": self.numPix**2,
                "scale": 1,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        kwargs_constraints = {"source_grid_offset": True}
        param_class = Param(
            kwargs_model,
            kwargs_fixed_source=kwargs_fixed_source,
            kwargs_fixed_lens_light=kwargs_fixed_lens_light,
            **kwargs_constraints
        )

        likelihood = LikelihoodModule(
            kwargs_data_joint=kwargs_data,
            kwargs_model=kwargs_model,
            param_class=param_class,
            **kwargs_likelihood
        )

        kwargs_source = [{"amp": np.ones(n_scales * numPix_source**2)}]
        kwargs_lens_light = [{"amp": np.ones(n_scales * self.numPix**2)}]
        kwargs_special = {"delta_x_source_grid": 0, "delta_y_source_grid": 0}
        args = param_class.kwargs2args(
            kwargs_lens=self.kwargs_lens,
            kwargs_source=kwargs_source,
            kwargs_lens_light=kwargs_lens_light,
            kwargs_special=kwargs_special,
        )

        logL = likelihood.logL(args, verbose=True)
        num_data_evaluate = likelihood.num_data
        npt.assert_almost_equal(logL / num_data_evaluate, -1 / 2.0, decimal=1)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import copy

import numpy.testing as npt
import numpy as np
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.Workflow.fitting_sequence import FittingSequence
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF


class TestAlignmentMatching(object):
    def setup(self):
        np.random.seed(41)
        # data specifics
        sigma_bkg = 0.01  # background noise per pixel
        exp_time = 1000  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 20  # cutout pixel size
        deltaPix = 0.2  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.1  # full width half max of PSF

        # PSF specification

        self.kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        transform_pix2angle = self.kwargs_data["transform_pix2angle"]
        self.kwargs_data2 = copy.deepcopy(self.kwargs_data)
        self.delta_x_offset = 0.2
        self.kwargs_data2["ra_at_xy_0"] += self.delta_x_offset
        self.phi_rot = 0.1
        cos_phi, sin_phi = np.cos(self.phi_rot), np.sin(self.phi_rot)
        rot_matrix = np.array([[cos_phi, -sin_phi], [sin_phi, cos_phi]])
        transform_pix2angle_rot = np.dot(transform_pix2angle, rot_matrix)
        self.kwargs_data2["transform_pix2angle"] = transform_pix2angle_rot

        data_class = ImageData(**self.kwargs_data)
        data_class2 = ImageData(**self.kwargs_data2)
        kwargs_psf = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "pixel_size": deltaPix,
            "truncation": 3,
        }
        psf_class = PSF(**kwargs_psf)
        kwargs_numerics = {"supersampling_factor": 1}

        self.kwargs_lens_light = [
            {
                "amp": 100.0,
                "R_sersic": 0.5,
                "n_sersic": 2,
                "e1": 0.3,
                "e2": -0.2,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        lens_light_model_list = ["SERSIC_ELLIPSE"]
        lens_light_model_class = LightModel(light_model_list=lens_light_model_list)

        imageModel = ImageModel(
            data_class, psf_class, lens_light_model_class=lens_light_model_class
        )
        imageModel2 = ImageModel(
            data_class2, psf_class, lens_light_model_class=lens_light_model_class
        )
        image_sim = sim_util.simulate_simple(
            imageModel, kwargs_lens_light=self.kwargs_lens_light
        )
        image_sim2 = sim_util.simulate_simple(
            imageModel2, kwargs_lens_light=self.kwargs_lens_light
        )
        self.kwargs_data["image_data"] = image_sim

        self.kwargs_data2_offset = copy.deepcopy(self.kwargs_data)
        self.kwargs_data2_offset["image_data"] = image_sim2

        image_band1 = [self.kwargs_data, kwargs_psf, kwargs_numerics]
        image_band2 = [self.kwargs_data2_offset, kwargs_psf, kwargs_numerics]
        multi_band_list = [image_band1, image_band2]
        self.kwargs_data_joint = {
            "multi_band_list": multi_band_list,
            "multi_band_type": "joint-linear",
        }

        self.kwargs_model = {"lens_light_model_list": lens_light_model_list}
        self.kwargs_constraints = {}
        self.kwargs_likelihood = {}
        lens_light_sigma = [
            {"R_sersic": 0.05, "n_sersic": 0.5, "center_x": 0.1, "center_y": 0.1}
        ]
        lens_light_lower = [
            {"R_sersic": 0.01, "n_sersic": 0.5, "center_x": -2, "center_y": -2}
        ]
        lens_light_upper = [
            {"R_sersic": 10, "n_sersic": 5.5, "center_x": 2, "center_y": 2}
        ]
        lens_light_param = (
            self.kwargs_lens_light,
            lens_light_sigma,
            [{}],
            lens_light_lower,
            lens_light_upper,
        )
        self.kwargs_params = {"lens_light_model": lens_light_param}

    def test_flux_calibration(self):
        fittingSequence = FittingSequence(
            self.kwargs_data_joint,
            self.kwargs_model,
            self.kwargs_constraints,
            self.kwargs_likelihood,
            self.kwargs_params,
        )
        fitting_list = []

        kwargs_align = {
            "n_particles": 20,
            "n_iterations": 40,
            "compute_bands": [False, True],
            "align_offset": True,
            "align_rotation": True,
            "delta_shift": 0.3,
            "delta_rot": 0.5,
        }

        fitting_list.append(["align_images", kwargs_align])
        chain_list = fittingSequence.fit_sequence(fitting_list)
        multi_band_list_new = fittingSequence.multi_band_list
        kwargs_data2_new = multi_band_list_new[1][0]
        ra_shift = kwargs_data2_new["ra_shift"]
        npt.assert_almost_equal(ra_shift, self.delta_x_offset, decimal=1)
        phi_rot = kwargs_data2_new["phi_rot"]
        npt.assert_almost_equal(phi_rot, self.phi_rot, decimal=1)



__author__ = "sibirrer"

import numpy.testing as npt
import numpy as np
import pytest
from lenstronomy.LensModel.Solver.epl_shear_solver import caustics_epl_shear
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.LensModel.lens_model import LensModel


class TestLensEquationSolver(object):
    def setup_method(self):
        """

        :return:
        """
        pass

    def test_spep_sis(self):
        lens_model_list = ["SPEP", "SIS"]
        lensModel = LensModel(lens_model_list)
        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = -0.1
        min_distance = 0.05
        search_window = 10
        gamma = 1.9
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": gamma,
                "e1": 0.2,
                "e2": -0.03,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"theta_E": 0.1, "center_x": 0.5, "center_y": 0},
        ]
        x_pos, y_pos = lensEquationSolver.image_position_from_source(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            min_distance=min_distance,
            search_window=search_window,
            precision_limit=10 ** (-10),
            num_iter_max=10,
        )
        source_x, source_y = lensModel.ray_shooting(x_pos, y_pos, kwargs_lens)
        npt.assert_almost_equal(sourcePos_x, source_x, decimal=10)

    def test_nfw(self):
        lens_model_list = ["NFW_ELLIPSE", "SIS"]
        lensModel = LensModel(lens_model_list)
        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = -0.1
        min_distance = 0.05
        search_window = 10
        Rs = 4.0
        kwargs_lens = [
            {
                "alpha_Rs": 1.0,
                "Rs": Rs,
                "e1": 0.2,
                "e2": -0.03,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"theta_E": 1, "center_x": 0, "center_y": 0},
        ]
        x_pos, y_pos = lensEquationSolver.image_position_from_source(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            min_distance=min_distance,
            search_window=search_window,
            precision_limit=10 ** (-10),
            num_iter_max=10,
            verbose=True,
            magnification_limit=1,
        )
        source_x, source_y = lensModel.ray_shooting(x_pos, y_pos, kwargs_lens)
        npt.assert_almost_equal(sourcePos_x, source_x, decimal=10)

    def test_multiplane(self):
        lens_model_list = ["SPEP", "SIS"]
        lensModel = LensModel(
            lens_model_list,
            z_source=1.0,
            lens_redshift_list=[0.5, 0.3],
            multi_plane=True,
        )
        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = -0.1
        min_distance = 0.05
        search_window = 10
        gamma = 1.9
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": gamma,
                "e1": 0.2,
                "e2": -0.03,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"theta_E": 0.1, "center_x": 0.5, "center_y": 0},
        ]
        x_pos, y_pos = lensEquationSolver.image_position_from_source(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            min_distance=min_distance,
            search_window=search_window,
            precision_limit=10 ** (-10),
            num_iter_max=10,
        )
        source_x, source_y = lensModel.ray_shooting(x_pos, y_pos, kwargs_lens)
        npt.assert_almost_equal(sourcePos_x, source_x, decimal=10)

    def test_central_image(self):
        lens_model_list = ["SPEP", "SIS", "SHEAR"]
        kwargs_spep = {
            "theta_E": 1,
            "gamma": 2,
            "e1": 0.2,
            "e2": -0.03,
            "center_x": 0,
            "center_y": 0,
        }
        kwargs_sis = {"theta_E": 1, "center_x": 1.5, "center_y": 0}
        kwargs_shear = {"gamma1": 0.01, "gamma2": 0}
        kwargs_lens = [kwargs_spep, kwargs_sis, kwargs_shear]
        lensModel = LensModel(lens_model_list)
        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = -0.1
        min_distance = 0.05
        search_window = 10
        x_pos, y_pos = lensEquationSolver.image_position_from_source(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            min_distance=min_distance,
            search_window=search_window,
            precision_limit=10 ** (-10),
            num_iter_max=10,
        )
        source_x, source_y = lensModel.ray_shooting(x_pos, y_pos, kwargs_lens)
        npt.assert_almost_equal(sourcePos_x, source_x, decimal=10)
        print(x_pos, y_pos)
        assert len(x_pos) == 4

    def test_example(self):
        lens_model_list = ["SPEP", "SHEAR"]
        lensModel = LensModel(lens_model_list)

        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.03
        sourcePos_y = 0.0
        min_distance = 0.05
        search_window = 10
        gamma = 2.0
        gamma1, gamma2 = -0.04, -0.1
        kwargs_shear = {
            "gamma1": gamma1,
            "gamma2": gamma2,
        }  # shear values to the source plane
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": gamma,
            "center_x": 0.0,
            "center_y": 0.0,
            "e1": 0.01,
            "e2": 0.05,
        }  # parameters of the deflector lens model

        kwargs_lens = [kwargs_spemd, kwargs_shear]
        x_pos, y_pos = lensEquationSolver.image_position_from_source(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            min_distance=min_distance,
            search_window=search_window,
            precision_limit=10 ** (-10),
            num_iter_max=10,
            arrival_time_sort=True,
        )

        (
            x_pos_non_linear,
            y_pos_non_linear,
        ) = lensEquationSolver.image_position_from_source(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            min_distance=min_distance,
            search_window=search_window,
            precision_limit=10 ** (-10),
            num_iter_max=10,
            arrival_time_sort=True,
            non_linear=True,
        )

        x_pos_stoch, y_pos_stoch = lensEquationSolver.image_position_from_source(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            solver="stochastic",
            search_window=search_window,
            precision_limit=10 ** (-10),
            arrival_time_sort=True,
            x_center=0,
            y_center=0,
            num_random=100,
        )
        assert len(x_pos) == 4
        assert len(x_pos_stoch) == 4
        assert len(x_pos_non_linear) == 4
        npt.assert_almost_equal(x_pos, x_pos_stoch, decimal=5)
        npt.assert_almost_equal(x_pos, x_pos_non_linear, decimal=5)

    def test_analytical_lens_equation_solver(self):
        lensModel = LensModel(["EPL_NUMBA", "SHEAR"])
        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.03
        sourcePos_y = 0.0
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": 2.2,
                "center_x": 0.01,
                "center_y": 0.02,
                "e1": 0.01,
                "e2": 0.05,
            },
            {"gamma1": -0.04, "gamma2": -0.1, "ra_0": 0.01, "dec_0": 0.02},
        ]

        x_pos, y_pos = lensEquationSolver.image_position_from_source(
            sourcePos_x, sourcePos_y, kwargs_lens, solver="analytical"
        )
        source_x, source_y = lensModel.ray_shooting(x_pos, y_pos, kwargs_lens)
        assert len(source_x) == len(source_y) >= 4
        npt.assert_almost_equal(sourcePos_x, source_x, decimal=10)
        npt.assert_almost_equal(sourcePos_y, source_y, decimal=10)

        x_pos_ls, y_pos_ls = lensEquationSolver.image_position_from_source(
            sourcePos_x, sourcePos_y, kwargs_lens, solver="analytical"
        )
        for x, y in zip(
            x_pos_ls, y_pos_ls
        ):  # Check if it found all solutions lenstronomy found
            assert np.sqrt((x - x_pos) ** 2 + (y - y_pos) ** 2).min() < 1e-8

        # here we test with shear and mass profile centroids not aligned
        lensModel = LensModel(["EPL_NUMBA", "SHEAR"])
        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.03
        sourcePos_y = 0.0
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": 2.2,
                "center_x": 0.01,
                "center_y": 0.02,
                "e1": 0.01,
                "e2": 0.05,
            },
            {"gamma1": -0.04, "gamma2": -0.1, "ra_0": 0.0, "dec_0": 1.0},
        ]

        x_pos, y_pos = lensEquationSolver.image_position_from_source(
            sourcePos_x, sourcePos_y, kwargs_lens, solver="analytical"
        )
        source_x, source_y = lensModel.ray_shooting(x_pos, y_pos, kwargs_lens)
        assert len(source_x) == len(source_y) >= 2
        npt.assert_almost_equal(sourcePos_x, source_x, decimal=10)
        npt.assert_almost_equal(sourcePos_y, source_y, decimal=10)

    def test_caustics(self):
        lm = LensModel(["EPL_NUMBA", "SHEAR"])
        leqs = LensEquationSolver(lm)

        kwargs = [
            {
                "theta_E": 1.0,
                "e1": 0.5,
                "e2": 0.1,
                "center_x": 0.0,
                "center_y": 0.0,
                "gamma": 1.9,
            },
            {"gamma1": 0.03, "gamma2": 0.01, "ra_0": 0.0, "dec_0": 0.0},
        ]

        # Calculate the caustics and a few critical curves.
        caus = caustics_epl_shear(kwargs, return_which="caustic")
        lensplane_caus = caustics_epl_shear(
            kwargs, return_which="caustic", sourceplane=False
        )
        cut = caustics_epl_shear(kwargs, return_which="cut")
        lensplane_cut = caustics_epl_shear(
            kwargs, return_which="cut", sourceplane=False
        )
        twoimg = caustics_epl_shear(kwargs, return_which="double")
        fourimg = caustics_epl_shear(kwargs, return_which="quad")
        min_mag = np.abs(lm.magnification(*lensplane_caus, kwargs)).min()
        assert min_mag > 1e12
        assert np.abs(lm.magnification(*lensplane_cut, kwargs)).min() > 1e12

        # Test whether the caustics indeed the number of images they say
        N = 20
        xpl, ypl = np.linspace(-1, 1, N), np.linspace(-1, 1, N)
        xgr, ygr = np.meshgrid(xpl, ypl, indexing="ij")
        xf, yf = xgr.flatten(), ygr.flatten()
        sols = [
            leqs.image_position_from_source(x, y, kwargs, solver="analytical")
            for x, y in zip(xf, yf)
        ]
        numsols = np.array([len(p[0]) for p in sols])

        from matplotlib.path import Path

        points = np.vstack((xf, yf)).T

        p = Path(twoimg.T)  # make a polygon
        grid2img = p.contains_points(points)
        assert np.all(numsols[grid2img] >= 2)

        p = Path(fourimg.T)  # make a polygon
        grid4img = p.contains_points(points)
        assert np.all(numsols[grid4img] >= 4)

    def test_analytical_sie(self):
        sourcePos_x = 0.03
        sourcePos_y = 0.0

        lensModel = LensModel(["SIE"])
        lensEquationSolver = LensEquationSolver(lensModel)
        kwargs_lens = [
            {"theta_E": 1.0, "center_x": 0.0, "center_y": 0.0, "e1": 0.5, "e2": 0.05},
        ]

        x_pos, y_pos = lensEquationSolver.image_position_from_source(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            solver="analytical",
            magnification_limit=1e-3,
        )
        source_x, source_y = lensModel.ray_shooting(x_pos, y_pos, kwargs_lens)
        assert len(source_x) == len(source_y) == 4
        npt.assert_almost_equal(sourcePos_x, source_x, decimal=10)
        npt.assert_almost_equal(sourcePos_y, source_y, decimal=10)

    def test_assertions(self):
        lensModel = LensModel(["SPEP"])
        lensEquationSolver = LensEquationSolver(lensModel)
        kwargs_lens = [
            {
                "theta_E": 1,
                "gamma": 2,
                "e1": 0.2,
                "e2": -0.03,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        with pytest.raises(ValueError):
            lensEquationSolver.image_position_from_source(
                0.1, 0.0, kwargs_lens, solver="analytical"
            )

        lensModel = LensModel(["EPL_NUMBA", "SHEAR"])
        lensEquationSolver = LensEquationSolver(lensModel)
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": 2.2,
                "center_x": 0.0,
                "center_y": 0.0,
                "e1": 0.01,
                "e2": 0.05,
            },
            {"gamma1": -0.04, "gamma2": -0.1, "ra_0": 0.0, "dec_0": 0.0},
        ]

        with pytest.raises(ValueError):
            lensEquationSolver.image_position_from_source(
                0.1, 0.0, kwargs_lens, solver="nonexisting"
            )


if __name__ == "__main__":
    pytest.main()
